{-# OPTIONS -fglasgow-exts -XOverlappingInstances -XIncoherentInstances #-}

module AST where

import qualified CST as CST
import Data.List
import Control.Monad.State
import Debug.Trace
import Data.Maybe
import Data.Graph
import Util
import Automaton

class ShowP a where
    showP :: a -> String

data PreExp a 
    = ECon'  String (PreExp a) (PreExp a)
    | EFun'  a [a]
    | EApp'  (PreExp a) (PreExp a)
    | EVar'  a 
    | ERec'  [(String,(PreExp a))]
    | EChar' Char
    | EEps'
    deriving (Eq,Ord)

data PrePat a b 
    = PCon'  String (PrePat a b) (PrePat a b)
    | PApp'  (PrePat a b) (PrePat a b)
    | PVar'  a b
    | PRec'  [(String, (PrePat a b))]
    | PChar' Char
    | PNil'
    deriving (Eq,Ord)


cst2preast :: CST.Prog String -> (G Int, [(String, [PrePat String Int], PreExp String)])
cst2preast decls = 
    let (fds, (_,_,tds)) = runState t (0,[],[])
    in (G tds, fds)
    where
      (tdecls, fdecls) = CST.divideDecl decls 
      t = do { f tdecls 
             ; fdecls' <- g fdecls 
             ; return fdecls' }
      f [] = return ()
      f (CST.TDecl x t:ts) = 
          do { i   <- uniqNameFrom x 
             ; t'  <- ctype2G t
             ; f ts
             ; addDeriv (i,t')
             ; return () }
      g [] = return []
      g (CST.FDecl f pats e:fs) = 
          do { pats' <- mapM cpat2ppat pats
             ; e'    <- cexp2pexp e
             ; fs'   <- g fs
             ; return $ (f,pats',e'):fs' }


uniqName = 
    do { (i,a,b) <- get
       ; put (i+1,a,b) 
       ; return i }

uniqNameFrom x =
    do { (i,a,b) <- get
       ; case lookup x a of 
           Just j -> return j 
           Nothing ->
               do { put (i+1,(x,i):a,b)
                  ; return i }}

addDeriv (x,tr) =
    do { (i,a,b) <- get
       ; put (i,a,(x,tr):b) }


ctype2G (CST.TApp t1 (CST.TNil)) = ctype2G t1 
ctype2G (CST.TApp (CST.TNil) t2) = ctype2G t2
ctype2G (CST.TApp (CST.TApp t1 t2) t3) = ctype2G (CST.TApp t1 (CST.TApp t2 t3))
ctype2G (CST.TCon c r t) =
    do { r' <- ctype2Gr r
       ; t' <- ctype2G t
       ; return $ DT c r' t' }
ctype2G (CST.TAnyLabel r t) =
    do { r' <- ctype2Gr r
       ; t' <- ctype2G  t
       ; return $ DAT r' t' }
ctype2G (CST.TApp t1 t2) =
    do { t1' <- ctype2G t1
       ; t2' <- ctype2G t2
       ; return $ DA t1' t2' }
ctype2G (CST.TSel t1 t2) = 
    do { t1' <- ctype2G t1
       ; t2' <- ctype2G t2
       ; i   <- uniqName 
       ; addDeriv (i,t1')
       ; addDeriv (i,t2')
       ; return $ (DS i) }
ctype2G (CST.TFun a) =
    do { i <- uniqNameFrom a 
       ; return $ (DS i) }
ctype2G (CST.TRec r) = ctype2Gr r
ctype2G (CST.TChar c) = return $ DC c 
ctype2G (CST.TAnyChar) = return $ DAnyC 
ctype2G (CST.TStar t) =
    do { i  <- uniqName 
       ; t' <- ctype2G t
       ; addDeriv (i,DA t' (DS i))
       ; addDeriv (i,DE)
       ; return $ DS i }
ctype2G (CST.TNil) = return $ DE
ctype2Gr (CST.TRFun a) =
    do { i <- uniqNameFrom a 
       ; return $ (DS i) }
ctype2Gr x =
    do (rs,ng) <- f x []
       return $ DR (sort rs) ng 
    where
      f (CST.TREmp)  _  = return ([], DNothing)
      f (CST.TRRest) ng = return ([], DWithout ng)
      f (CST.TRCons (k,t) rs) ng =
          do { t' <- ctype2G t
             ; (rs',ng') <- f rs (k:ng)
             ; return $ ((k,t'):rs', ng') }

cpat2ppat (CST.PCon c r p) = 
    do { r' <- cpat2ppat' r
       ; p' <- cpat2ppat  p 
       ; return $ PCon' c (PRec' r') p' }
cpat2ppat (CST.PApp p1 p2) =
    do { p1' <- cpat2ppat p1 
       ; p2' <- cpat2ppat p2 
       ; return $ PApp' p1' p2'  }
cpat2ppat (CST.PVar v t) =
    do { t' <- ctype2G t
       ; case t' of 
           DS j -> 
               return $ PVar' v j 
           _ ->
               do { i  <- uniqName 
                  ; addDeriv (i,t')
                  ; return $ PVar' v i }}
cpat2ppat (CST.PLit []) = cpat2ppat (CST.PNil)
cpat2ppat (CST.PLit s) =
    return $ foldr1 PApp' $ map PChar' s
cpat2ppat (CST.PNil) = return $ PNil'
cpat2ppat' (CST.REmp) = return []
cpat2ppat' (CST.RCons (k,p) r) =
    do { r' <- cpat2ppat' r
       ; p' <- cpat2ppat  p
       ; return $ (k,p'):r' }


cexp2pexp (CST.ECon c r e) =
    do { r' <- cexp2pexp r
       ; e' <- cexp2pexp e
       ; return $ ECon' c r' e' }
cexp2pexp (CST.EFun f vs) = return $ EFun' f vs
cexp2pexp (CST.EApp e1 e2) = 
    do { e1' <- cexp2pexp e1
       ; e2' <- cexp2pexp e2
       ; return $ EApp' e1' e2' }
cexp2pexp (CST.EVar x) = return $ EVar' x
cexp2pexp (CST.ELit []) = return $ EEps' 
cexp2pexp (CST.ELit s) =
    return $ foldr1 EApp' $ map EChar' s
cexp2pexp (CST.ERec rs) =
    do { rs' <-  mapM (\(k,e)-> do { e' <- cexp2pexp e; return (k,e') }) rs 
       ; return $ ERec' rs' }
cexp2pexp (CST.ENil) = return $ EEps'





-- | a: Variable and Function name, b: Identifier
data IdExp a b 
    = ECon b String (IdExp a b) (IdExp a b)
    | EFun b a [a]
    | EApp b (IdExp a b) (IdExp a b)
    | EVar b a -- First represents type variable.
    | ERec b [(String,(IdExp a b))]
    | EChar b Char
    | EEps b 
    | EFCon b (String,Int) [IdExp a b]                   -- ^ For Complement Function 
    | ELen  b b (IdExp a b)                              -- ^ For Complement Function 
    | ESplitAt b a a -- ^ For Inversed Tupled Function
      deriving (Eq,Ord)

expId (ECon i _ _ _) = i
expId (EFun i _ _)   = i
expId (EApp i _ _)   = i
expId (EVar i _)     = i
expId (ERec i _)     = i
expId (EChar i _)    = i
expId (EEps  i)      = i


                
instance (Show b) => Show (IdExp String b) where
    show (ECon i c r e)      = "#"++show i ++ "<"++c++">"++show r ++ "("++show e ++ ")"
    show (EFun i f vs)       = "#"++show i ++ f ++ "("++implode "," (map show vs)++")"
    show (EApp i e1 e2)      = "#"++show i ++ "(" ++ show e1 ++ "." ++ show e2 ++ ")"
    show (EVar i v)          = "#"++show i ++ v
    show (ERec i rs)         = "#"++show i ++ "{" ++ implode "," (map (\(k,t) -> k++":"++show t) rs) ++ "}"
    show (EChar i c)         = "#"++show i ++ "'" ++ [c] ++ "'"
    show (EEps i)            = "#"++show i ++ "$"
    show (EFCon i (r,j) es)  = "#"++show i++r++"_"++show j++ show es
    show (ELen  i j e)       = "#"++show i++"length"++show j ++"("++show e++")"
    show (ESplitAt i v1 v2)  = "#"++show i++"splitAt("++v1++","++v2++")"


instance (Show a, Show b) => Show (IdExp a b) where
    show (ECon i c r e) = "#"++show i ++ "<"++c++">"++show r ++ "("++show e ++ ")"
    show (EFun i f vs)  = "#"++show i ++ show f ++ "("++implode "," (map show vs)++")"
    show (EApp i e1 e2) = "#"++show i ++ "(" ++ show e1 ++ "." ++ show e2 ++ ")"
    show (EVar i v)     = "#"++show i ++ show v
    show (ERec i rs)    = "#"++show i ++ "{" ++ implode "," (map (\(k,t) -> k++":"++show t) rs) ++ "}"
    show (EChar i c)    = "#"++show i ++ "'" ++ [c] ++ "'"
    show (EEps i)       = "#"++show i ++ "$"
    show (EFCon i (r,j) es)  = "#"++show i++r++"_"++show j++ show es
    show (ELen  i j e)       = "#"++show i++"length"++show j++"("++show e++")"
    show (ESplitAt i v1 v2)  = "#"++show i++"splitAt("++show v1++","++show v2++")"

instance ShowP a => ShowP [a] where
    showP xs = "[" ++ implode "," (map showP xs) ++ "]"

instance Show b => ShowP (IdExp String b) where
    showP (ECon i c (ERec _ []) (EEps _)) = "<"++c++">"
    showP (ECon i c (ERec _ []) e)        = "<"++c++">"++"("++showP e++")"
    showP (ECon i c r e)                  = "<"++c++">"++showP r ++ "("++showP e ++ ")"
    showP (EFun i f vs)       = f ++ "("++implode "," (map show vs)++")"
    showP (EApp i e1 e2)      = showP e1 ++ "." ++ showP e2
    showP (EVar i v)          = v
    showP (ERec i rs)         = "{" ++ implode "," (map (\(k,t) -> k++":"++showP t) rs) ++ "}"
    showP (EChar i c)         = "'" ++ [c] ++ "'"
    showP (EEps i)            = "$"
    showP (EFCon i (r,j) es)  = r++"_"++show j++ showP es
    showP (ELen  i  j e)      = "length("++showP e++")"
    showP (ESplitAt i v1 v2)  = "splitAt("++v1++","++v2++")"

instance (Show a,Show b) => ShowP (IdExp a b) where
    showP (ECon i c (ERec _ []) (EEps _)) = "<"++c++">"
    showP (ECon i c (ERec _ []) e)        = "<"++c++">"++"("++showP e++")"
    showP (ECon i c r e)      = "<"++c++">"++showP r ++ "("++showP e ++ ")"
    showP (EFun i f vs)       = show f ++ "("++implode "," (map show vs)++")"
    showP (EApp i e1 e2)      = showP e1 ++ "." ++ showP e2
    showP (EVar i v)          = show v
    showP (ERec i rs)         = "{" ++ implode "," (map (\(k,t) -> k++":"++showP t) rs) ++ "}"
    showP (EChar i c)         = "'" ++ [c] ++ "'"
    showP (EEps i)            = "$"
    showP (EFCon i (r,j) es)  = r++"_"++show j++ showP es
    showP (ELen  i j e)       = "length("++showP e++")"
    showP (ESplitAt i v1 v2)  = "splitAt("++show v1++","++show v2++")"

                                
-- | a; Variable Name, b: Identifier
data IdPat a b 
    = PCon  b String (IdPat a b) (IdPat a b)
    | PApp  b (IdPat a b) (IdPat a b)
    | PVar  b a 
    | PChar b Char
    | PRec  b [(String, (IdPat a b))]
    | PNil  b
    | PFCon b (String,Int) [IdPat a b]               -- ^ For Inversed Function
      deriving (Eq, Ord)

patId (PCon i _ _ _) = i
patId (PApp i _ _)   = i
patId (PVar i _)     = i
patId (PRec i _)     = i
patId (PChar i _)    = i
patId (PNil  i)      = i


instance (Show a, Show b) => Show (IdPat a b) where
    show (PCon i c r e) = "#"++show i++ "<"++c++">"++show r++"("++show e++")"
    show (PApp i e1 e2) = "#"++show i++ "("++ show e1 ++ " . " ++show e2 ++")"
    show (PVar i a)     = "#"++show i++show a
    show (PChar i c)    = "#"++show i++['"',c,'"']
    show (PRec i rs)    = "#"++show i ++ "{" ++ implode "," (map (\(k,t) -> k++":"++show t) rs) ++ "}"
    show (PNil i)       = "#"++show i++"$"
    show (PFCon i (r,j) es)  = "#"++show i++r++"_"++show j++ show es

instance Show b => ShowP (IdPat String b) where
    showP (PCon i c (PRec _ []) (PNil _)) = "<"++c++">"
    showP (PCon i c (PRec _ []) e)        = "<"++c++">"++"("++showP e++")"
    showP (PCon i c r e) = "<"++c++">"++showP r++"("++showP e++")"
    showP (PApp i e1 e2) = showP e1 ++ " . " ++showP e2
    showP (PVar i a)     = a ++ " :: " ++ "Type"++show i
    showP (PChar i c)    = ['\'',c,'\'']
    showP (PRec i rs)    = "{" ++ implode "," (map (\(k,t) -> k++":"++showP t) rs) ++ "}"
    showP (PNil i)       = "$"
    showP (PFCon i (r,j) es)  = show i++r++"_"++show j++ show es

instance (Show a, Show b) => ShowP (IdPat a b) where
    showP (PCon i c (PRec _ []) (PNil _)) = "<"++c++">"
    showP (PCon i c (PRec _ []) e)        = "<"++c++">"++"("++showP e++")"
    showP (PCon i c r e) = "<"++c++">"++showP r++"("++showP e++")"
    showP (PApp i e1 e2) = showP e1 ++ " . " ++showP e2
    showP (PVar i a)     = show a ++ " :: " ++ "Type"++show i
    showP (PChar i c)    = ['\"',c,'\"']
    showP (PRec i rs)    = "{" ++ implode "," (map (\(k,t) -> k++":"++showP t) rs) ++ "}"
    showP (PNil i)       = "$"
    showP (PFCon i (r,j) es)  = show i++r++"_"++show j++ show es
                           
                                                                                 
newtype P a b = P [(Int, (a,[IdPat a b], IdExp a b))]
-- ^ Rule Id, (Function Name, List of IDed Patterns, IDed Expressioon)

instance Show b =>Show (P String b) where
    show (P xs) = implode "\n" (map s xs)
        where s (i,(f,pats,e)) = show i++":\n"++f++"("++implode "," (map show pats) ++ ")" ++ " = " ++ "\n" ++
                                 "     " ++ show e
instance (Show a, Show b) => Show (P a b) where
    show (P xs) = implode "\n" (map s xs)
        where s (i,(f,pats,e)) = show i ++ ":\n" ++ show f ++ "(" ++ implode "," (map show pats) ++ ")" ++ " = " ++ "\n" ++
                                 "     " ++ show e

instance (Show b) => ShowP (P String b) where
    showP (P xs) = implode "\n" (map s xs)
        where s (i,(f,pats,e)) = show i++":\n"++f++"("++implode "," (map showP pats) ++ ")" ++ " = " ++ "\n" ++
                                 "     " ++ showP e

instance (Show a,Show b) => ShowP (P a b) where
    showP (P xs) = implode "\n" (map s xs)
        where s (i,(f,pats,e)) = show i++":\n"++show f++"("++implode "," (map showP pats) ++ ")" ++ " = " ++ "\n" ++
                                 "     " ++ showP e

importantStates :: (Ord b, Eq b) => P a b -> [b]
importantStates (P rules) =
    snub $ concatMap h rules
    where
      h (_,(_,ps,_)) = concatMap f ps
      f (PCon _ _ r e) = f r ++ f e
      f (PApp _ e1 e2) = f e1 ++ f e2
      f (PVar i _)     = [i] 
      f (PChar _ _)    = []
      f (PRec _ rs)    = concatMap (f.snd) rs
      f (PNil _)       = []
      f (PFCon _ _ es) = concatMap f es

-- | Vartically-Regular and Horizontary-Context-Free Tree Grammar
newtype G a = G { derivations:: [(a,Deriv a)] }
data Deriv a 
    = DT String (Deriv a) (Deriv a)
    | DC Char 
    | DS a
    | DAT (Deriv a) (Deriv a)
    | DAnyC
    | DA (Deriv a) (Deriv a)
    | DR [(String, Deriv a)] DNG
    | DE
      deriving (Eq,Ord)

instance Functor G where
    fmap f (G x) = G $ map (\(x,y) -> (f x, fmap f y)) x

instance Functor Deriv where
    fmap f (DT c r e) = DT c (fmap f r) (fmap f e)
    fmap f (DC c)     = DC c
    fmap f (DS a)     = DS $ f a
    fmap f (DAT r e)  = DAT (fmap f r) (fmap f e)
    fmap f DAnyC      = DAnyC
    fmap f (DA t1 t2) = DA (fmap f t1) (fmap f t2)
    fmap f (DR rs ng) = DR (map (\(k,t) -> (k,fmap f t)) rs) ng
    fmap f DE         = DE


instance Show a => Show (Deriv a) where
    show (DT c r e) = "<"++c++">"++show r ++ "("++show e ++")"
    show (DC c)     = ['#',c]
    show (DS a)     = show a
    show (DAT r e)  = "~"++show r++"("++show e++")"
    show (DAnyC)    = "_"
    show (DA t1 t2) = show t1 ++ " . " ++ show t2
    show (DR rs ng) = "{"++ implode "," (map (\(k,t) -> k++":"++show t) rs) ++ show ng++"}" 
    show DE         = "$"

instance Show a => Show (G a) where
    show (G x) = showRed x

data DNG = DNothing
         | DWithout [String]
           deriving (Eq,Ord,Show)

variables :: (Ord a,Eq a) => G a -> [a]
variables (G xs) = snub $ concatMap (\(x,tr) -> x:vars tr) xs

vars :: Deriv a -> [a]
vars (DT c r e) = vars r ++ vars e
vars (DS a)     = [a]
vars (DA t1 t2) = vars t1 ++ vars t2
vars (DR rs _)  = concatMap (vars.snd) rs
vars _          = []

{- | 
   Automaton State for a Program.
   SF: a state for a function name.
   SE: a state for an expression.
-}
data AState a b = SF a | SF' a | SE b | UD
                deriving (Eq,Ord)

instance (Show b) => Show (AState String b) where
    show (SF a) = a
    show (SF' a) = a++"'"
    show (SE b) = show b
    show UD     = ""

instance (Show a, Show b) => Show (AState a b) where
    show (SF a) = show a
    show (SF' a) = show a ++ "'"
    show (SE b) = show b
    show UD     = ""

{-|
 Mohri & Nederhof's algorithm to approximate 
 context-free language decribed by CFG using regular grammar.

 We use the tag Left for the original states and 
 the states representing "prefix" in their algorithm,
 and use the tag Right for the states representing "postfix" in the algorithm.
 
 For example, for the input CFG
  A --> a A b | c
 This implementation returns the following RG
  A --> a A, A' -> b A', A -> c A', A' -> 
-}
-- mn_transform :: (Eq a, Ord a) => G a -> G (Either a a)
mn_transform (G derivs) = 
    G $ snub $ end_nils $ concatMap (\(vs,drs)->check vs drs) (part_deriv mutus derivs)
    where mutus = mutually_recursive_names (G derivs)          
          end_nils xs = xs ++ concatMap (\x -> case x of
                                                 Right y -> [(x,DE)]
                                                 _ -> []) (AST.variables (G xs))
          allRightLinear vs drs = all (\(x,tr) -> f tr) drs
              where f (DT c r e) = f r && f e
                    f (DAT  r e) = f r && f e
                    f (DA t1 DE) = f t1
                    f (DA DE t2) = f t2
                    f (DA (DA t1 t2) t3) = f (DA t1 (DA t2 t3))
                    f (DA (DS a) _) = not $ a `elem` vs
                    f (DA t1 t2) = f t1 && f t2
                    f (DR rs _)  = all (f.snd) rs
                    f _          = True
          mp f = derivations . fmap f . G
          part_deriv [] [] = []
          part_deriv (vs:vss) drs = 
              let (drs1,drs2) = partition (\(x,tr)->x `elem` vs) drs
              in (vs,drs1):part_deriv vss drs2
          check vs drs = 
              if allRightLinear vs drs then
                  mp Left drs
              else
                  concatMap (transform vs) drs
          transform vs (x,tr) = 
              case runState (splitT tr) ([],[]) of 
                (tr', (drs,[]))   -> (Left x, da (fmap Left tr') (DS $ Right x)):concatMap trans drs
                (tr', (drs,[dr])) -> (Left x, fmap Left tr'):transT dr++concatMap trans drs
              where
                da DE  t = t
                da t1 t2 = DA t1 t2
                transT (y,tr) = 
                    case runState (splitT tr) ([],[]) of
                      (tr', (drs,[]))   -> (Right y, da (fmap Left tr') (DS $ Right x)):concatMap trans drs
                      (tr', (drs,[dr])) -> (Right y, fmap Left tr'):transT dr++concatMap trans drs
                trans (y,tr) =
                    case runState (split tr) ([],[]) of 
                      (tr', (drs1,[])) -> (Right y, fmap Left tr'):concatMap trans (drs1)
                add  (a,t) = do { (t1,t2) <- get; put ((a,t):t1,t2) }
                addT (a,t) = do { (t1,t2) <- get; put (t1,(a,t):t2) }
                splitT (DT c r e) =
                    do r' <- split r
                       e' <- split e
                       return $ DT c r' e'
                splitT (DAT r e) = 
                    do r' <- split r
                       e' <- split e 
                       return $ DAT r' e'
                splitT (DA (DA t1 t2) t3) = splitT (DA t1 (DA t2 t3))
                splitT (DA (DS a) t) | a `elem` vs = do { addT (a,t); return $ DS a }
                                     | otherwise   = do { t' <- splitT t; return $ DA (DS a) t' }
                splitT (DA t1 t2) = 
                    do t1' <- split  t1
                       t2' <- splitT t2
                       return $ DA t1' t2'
                splitT (DR rs ng) =
                    do ts' <- mapM split (map snd rs)
                       return $ DR (zip (map fst rs) ts' ) ng
                splitT (DS a) | a `elem` vs = do { addT (a,DE); return $ DS a }
                splitT t = return t
                split (DAT r e) = 
                    do r' <- split r
                       e' <- split e 
                       return $ DAT r' e'
                split (DT c r e) =
                    do r' <- split r
                       e' <- split e
                       return $ DT c r' e'
                split (DA (DA t1 t2) t3) = split (DA t1 (DA t1 t2))
                split (DA (DS a) t) | a `elem` vs = do { add (a,t); return $ DS a }
                                    | otherwise   = do { t' <- split t; return $ DA (DS a) t' }
                split (DA t1 t2) = 
                    do t1' <- split  t1
                       t2' <- split t2
                       return $ DA t1' t2'
                split (DR rs ng) =
                    do ts' <- mapM split (map snd rs)
                       return $ DR (zip (map fst rs) ts' ) ng
                split t = return t


mutually_recursive_names :: (Eq a, Ord a) => G a -> [[a]]
mutually_recursive_names (G derives) = 
    sccs2llist sccs
    where
      rels = mg $ snub $ map (\(x,y) -> (x,x,vars y)) derives
      mg []  = []
      mg [x] = [x]
      mg ((x,_,xs):(y,_,ys):r) | x == y = mg ((x,x,ys++xs):r)
                               | x /= y = (x,x,xs):mg ((y,y,ys):r)
      sccs = stronglyConnComp rels
      sccs2llist [] = []
      sccs2llist (AcyclicSCC v:xs) = [v]:sccs2llist xs
      sccs2llist (CyclicSCC x:xs)  = x:sccs2llist xs


statify :: P a Int -> Automaton Int -> Automaton (AState a Int)
statify (P program) (Automaton am) = 
    Automaton $ (transitions $ fmap SE (Automaton am)) ++ tr program
        where
          tr [] = []
          tr ((i,(f,ps,e)):decls) = (SF f, TN $ SE (expId e)):tr decls

-- regular_approximate :: (Ord a, Eq a) => G (AState a Int) -> Automaton (AState a Int)
regular_approximate (G transes) = -- trace (show $ G transes)$ 
    eliminate_epsrules $ unifyToEmp $ unifyToNil $ tvar $ a
    where
      (exp_rules, fun_rules) = partition (\(x,_) -> case x of { SE _ -> True; _ -> False }) transes
      G fun_rules'           = mn_transform (G fun_rules)
      G exp_rules'           = fmap Left (G exp_rules)
      a = c2r (G $ fun_rules' ++ exp_rules')
      vs = Automaton.states a
      lmax = maximum $ 0:map (\x -> case x of Left (Left (SE i)) -> i 
                                              _ -> 0) vs 
      tvar = fmap f 
          where f (Left (Left (SE i))) = SE i
                f (Left (Left (SF g)))  = SF g
                f (Left (Right (SF g))) = SF' g
                f (Right i)             = SE $ i+lmax + 1
      skip n [] = []
      skip n ((m,t):x) | n == m = skip n x
                       | n /= m = (m,t):x
      chk (SE i) = i > lmax+1
      chk _      = False
      unifyToNil (Automaton t) = 
          let (n1,n2) = partition chk toNilSingleRules 
              m = head (n2++n1++[SE 0]) -- 0: banpei
              sts = map (\n -> (n,m)) n1
              nn  = n1 \\ [m]
          in Automaton $ filter (\(x,y) -> not(x `elem` nn))  $ renameTranses' sts t
              where 
                toNilSingleRules = f t
                f [] = []
                f [(n, TEps)] = [n]
                f [x]         = []
                f ((n, TEps):(m,t):x) = 
                    if n /= m then
                        n:f ((m,t):x)
                    else
                        f (skip n x)
                f ((n,t):y:tr) = f (skip n (y:tr))
      unifyToEmp (Automaton t) = 
          let (n1,n2) = partition chk toEmpSingleRules 
              m = head (n2++n1++[SE 0]) -- 0: banpei
              sts = map (\n -> (n,m)) n1
              nn  = n1 \\ [m]
          in Automaton $ filter (\(x,y) -> not(x `elem` nn))  $ renameTranses' sts t
              where 
                toEmpSingleRules = f t
                f [] = []
                f [(n, TR [] RNothing)] = [n]
                f [x]         = []
                f ((n, TR [] RNothing):(m,t):x) = 
                    if n /= m then
                        n:f ((m,t):x)
                    else
                        f (skip n x)
                f ((n,t):y:tr) = f (skip n (y:tr))

-- {-|
-- -}
-- -- regular_approximate :: G Int -> Automaton Int
-- regular_approximate :: G Int -> P String Int -> Automaton (AState String Int)
-- regular_approximate g program = eliminate_epsrules $ statify program $ (\(Automaton s)-> Automaton $ snub s) $ unifyToEmp $ unifyToNil $ tvar $ a
--     where 
--       g' = mn_transform g
--       a  = c2r g'
--       vs = Automaton.states a
--       llmax = maximum $ 0:map (\x -> case x of Left (Left i) -> i 
--                                                _ -> 0) vs 
--       lrmax = maximum $ 0:map (\x -> case x of Left (Right i) -> i 
--                                                _ -> 0) vs
--       tvar = fmap f 
--           where f (Left (Left i)) = i
--                 f (Left (Right i)) = i+llmax+1
--                 f (Right i)        = i+llmax+lrmax+2
--       skip n [] = []
--       skip n ((m,t):x) | n == m = skip n x
--                        | n /= m = (m,t):x
--       unifyToNil (Automaton t) = 
--           let (n1,n2) = partition (>llmax+1) toNilSingleRules 
--               m = head (n2++n1++[0]) -- 0: banpei
--               sts = map (\n -> (n,m)) n1
--               nn  = n1 \\ [m]
--           in Automaton $ filter (\(x,y) -> not(x `elem` nn))  $ renameTranses' sts t
--               where 
--                 toNilSingleRules = f t
--                 f [] = []
--                 f [(n, TEps)] = [n]
--                 f [x]         = []
--                 f ((n, TEps):(m,t):x) = 
--                     if n /= m then
--                         n:f ((m,t):x)
--                     else
--                         f (skip n x)
--                 f ((n,t):y:tr) = f (skip n (y:tr))
--       unifyToEmp (Automaton t) = 
--           let (n1,n2) = partition (>llmax+1) toEmpSingleRules 
--               m = head (n2++n1++[0]) -- 0: banpei
--               sts = map (\n -> (n,m)) n1
--               nn  = n1 \\ [m]
--           in Automaton $ filter (\(x,y) -> not(x `elem` nn))  $ renameTranses' sts t
--               where 
--                 toEmpSingleRules = f t
--                 f [] = []
--                 f [(n, TR [] RNothing)] = [n]
--                 f [x]         = []
--                 f ((n, TR [] RNothing):(m,t):x) = 
--                     if n /= m then
--                         n:f ((m,t):x)
--                     else
--                         f (skip n x)
--                 f ((n,t):y:tr) = f (skip n (y:tr))


{-|
This function convert strongly-regular CFGs to RGs.
In strongly-regular CFG, the mutual recursive derivation 
noterminals appear only in tail positions.

It is worth noting that if the original CFG is not strongly regular.
This function may not terminate.
-}
-- c2r :: Eq a =>G a -> Automaton (Either a Int)
c2r (G derivs) = -- trace (show (G derivs)) $ 
    eliminate_epsrules $ Automaton $ snub transes
    where 
      transes = let (_,_,td) = execState (fs derivs) (0,[],[]) in td
      fs [] = return ()
      fs ((x,tr):xs) = 
          do f (Left x) tr
             fs xs 
      uniq' = do { (i,ct,td) <- get; put (i+1,ct,td); return $Right i }
      uniq r = case r of { DS a -> return (Left a); _ -> uniq' }
      add t =  do { (i,ct,td) <- get; put (i,ct,t:td) }
      putCtxt c n = do { (i,ct,td) <- get; put (i,(c,n):ct,td) }
      lookupCtxt v = do { (i,ct,td) <- get; return $ lookup v ct }
      expand a = map snd $ filter (\(x,t) -> x == a) derivs
      f x DE = add (x,TEps)
      f x (DS a) | x == Left a = return ()
      f x (DS a) = add (x,TN (Left a))
      f x (DR rs ng) = 
          do ns <- mapM (\t -> do { n <- uniq t; f n t; return n }) (map snd rs)
             add (x,TR (zip (map fst rs) ns) ng')
          where ng' = case ng of {DWithout r -> RWithout r; DNothing -> RNothing }
      f x (DA (DA t1 t2) t3) = f x (DA t1 (DA t2 t3)) 
      f x (DA DE t) = f x t
      f x (DA (DS a) DE) = f x (DS a)
      f x (DA (DC c) t1) = 
          do n1 <- uniq t1
             f n1 t1
             add (x,TT (NChar c) [n1])
      f x (DA (DT c r t1) t2) =
          do n1 <- uniq r
             n2 <- uniq t1
             n3 <- uniq t2
             f n1 r
             f n2 t1
             f n3 t2
             add (x,TT (NLabel c) [n1,n2,n3])
      f x (DA (DAT r t1) t2) = 
          do n1 <- uniq r
             n2 <- uniq t1
             n3 <- uniq t2
             f n1 r
             f n2 t1
             f n3 t2
             add (x,TT AnyLabel [n1,n2,n3])
--       f x (DA DAnyT t1) =
--           do n1 <- uniq t1
--              f n1 t1
--              r <- uniq'             
--              add (r, TR [] RNothing) -- FIXME
--              y <- uniq'
--              add (y, TT (AnyLabel) [r,y,y])
--              add (y, TEps)
--              add (x, TT (AnyLabel) [r,y,n1])
      f x (DA DAnyC t1) =
          do n1 <- uniq t1
             f n1 t1
             add (x, TT (AnyChar) [n1]) 
      f x (DA (DS a) t) = -- trace (show (DA (DS a) t)) $ 
          do v <- lookupCtxt (DA (DS a) t)
             case v of
               Just i -> add (x,TN i)
               Nothing ->
                   do putCtxt (DA (DS a) t) x
                      mapM_ (\t' -> f x (DA t' t)) (expand a)
      f x (DA t DE) = f x t
      f x t = f x (DA t DE)

          

cst2ast decls = preast2ast $ cst2preast (defs ++ decls) --numerize (defs++decls)
    where
      defs =
          [ CST.TDecl "Any" (CST.TSel (CST.TAnyLabel (CST.TRRest) (CST.TFun "Any") `CST.TApp` CST.TFun "Any") (CST.TNil)),
            CST.TDecl "String" (CST.TSel (CST.TAnyChar `CST.TApp` CST.TFun "String") (CST.TNil))]


toG decls (_,p,e) = f e ++ concatMap g p
    where      
      bind = concatMap binds p
      f (ECon i c r e) = (i, DT c (DS$expId r) (DS$expId e)):f r++f e
      f (EFun i g _)   = [ (i,DS$expId e') | (g',_,e') <- decls, g == g' ]
      f (EApp i e e')  = (i, DA (DS$expId e) (DS$expId e')):f e++f e'
      f (EVar i v)     = [(i,DS$lookup' v bind )]
      f (ERec i rs)    = (i, DR (map (\(k,t)->(k,DS$expId t)) rs) DNothing):concatMap (f.snd) rs
      f (EChar i c)    = [(i, DC c)]
      f (EEps i)       = [(i,DE)]
      g (PCon i c r p) = (i, DT c (DS$patId r) (DS$patId p)):g r++g p
      g (PApp i p p')  = (i, DA (DS$patId p) (DS$patId p')):g p++g p'
      g (PRec i rs)    = (i, DR (map (\(k,t)->(k,DS$patId t)) rs) DNothing):concatMap (g.snd) rs
      g (PVar i v)     = []
      g (PChar i c)    = [(i,DC c)]
      g (PNil i)       = [(i,DE)]

binds (PCon  i c r p) = bindsR r ++binds p
binds (PApp  i p1 p2) = binds p1 ++ binds p2
binds (PVar  i v1)    = [(v1,i)]
binds (PChar i c)     = []
binds (PNil  i)       = []
bindsR (PRec i rs) = bindsR' rs
    where bindsR' ((k,p):r)      = binds p ++ bindsR' r
          bindsR' ([])           = []

toGrammar (_,_,e) = f e 
    where
      f (ECon i c r e) = (SE i, DT c (DS$h r) (DS$h e)):f r++f e
      f (EFun i g _)   = [(SE i, DS $ SF g)]
      f (EApp i e e')  = (SE i, DA (DS$h e) (DS$h e')):f e++f e'
      f (EVar i v)     = []
      f (ERec i rs)    = (SE i, DR (map (\(k,t)->(k,DS$h t)) rs) DNothing):concatMap (f.snd) rs
      f (EChar i c)    = [(SE i, DC c)]
      f (EEps i)       = [(SE i,DE)]
      h = SE . expId 


preast2ast :: (Ord a, Eq a) => (G Int, [(a, [PrePat a Int], PreExp a)]) -> (G (AState a Int), P a Int)
preast2ast (G tdecls, fdecls) = 
    let fdecls'   = evalState (mapM t fdecls) (1 + (maximum ((-1):variables (G tdecls))))
        G tdecls' = fmap SE (G tdecls) 
    in (G (snub (tdecls' ++ concatMap toGrammar fdecls' ++ map fgram fdecls')), P $ zip [1..] fdecls')
    where
      fgram (f,_,e) = (SF f, convT e) 
          where 
            convT (ECon _ c r e) = DT c (convT r) (convT e)
            convT (EApp _ e1 e2) = DA (convT e1) (convT e2)
            convT (EFun _ g _)   = DS (SF g)
            convT (EVar i _)     = DS (SE i)
            convT (ERec _ rs)    = DR (map (\(k,t)->(k,convT t)) rs) DNothing
            convT (EChar _ c)    = DC c 
            convT (EEps _)       = DE
      t (f,pats,e) =
          do { pats' <- mapM introIDp pats
             ; e'    <- introIDe (concatMap binds pats') e
             ; return (f,pats',e') }
      introIDp (PCon' c r p) = 
          do { i  <- uniq
             ; r' <- introIDp r
             ; p' <- introIDp p
             ; return $ PCon i c r' p' }
      introIDp (PApp' p1 p2) =
          do { i <- uniq
             ; p1' <- introIDp p1
             ; p2' <- introIDp p2
             ; return $ PApp i p1' p2' }
      introIDp (PVar' v i) = return $ PVar i v 
      introIDp (PRec' rs)  =
          do { i <- uniq 
             ; rs' <- mapM (\(k,t) -> do { t' <- introIDp t; return (k,t') } ) rs             
             ; return $ PRec i rs' }
      introIDp (PChar' c) = 
          do { i <- uniq
             ; return $ PChar i c }
      introIDp (PNil') =
          do { i <- uniq
             ; return $ PNil i }
      introIDe binds (ECon' c r e) =
          do { i  <- uniq
             ; r' <- introIDe binds r 
             ; e' <- introIDe binds e
             ; return $ ECon i c r' e' }
      introIDe binds (EApp' e1 e2) =
          do { i <- uniq
             ; e1' <- introIDe binds e1 
             ; e2' <- introIDe binds e2
             ; return $ EApp i e1' e2' }
      introIDe binds (EFun' f vs) =
          do { i <- uniq
             ; return $ EFun i f vs }
      introIDe binds (EVar' v) =
          return $ EVar (lookup' v binds) v
      introIDe binds (ERec' rs) =
          do { i <- uniq 
             ; rs' <- mapM (\(k,t) -> do { t' <- introIDe binds t; return (k,t') } ) rs             
             ; return $ ERec i rs' }
      introIDe binds (EChar' c) =
          do { i <- uniq
             ; return $ EChar i c }
      introIDe binds (EEps') =
          do { i <- uniq
             ; return $ EEps i }
      uniq = do { i <- get; put (i+1); return i }
      lookup' k xs = case lookup k xs of
                       Just v -> v
                       Nothing -> error $ "``lookup'' failed at preast2ast." 


{-
 will be removed 
-}
numerize decls = evalState t (0,[])
    where
      (tdecls, fdecls) = CST.divideDecl decls
      t = do (tdecls',td)  <- f tdecls []
             (fdecls',td') <- h fdecls (td++tdecls')
             return $ (G (td'++concatMap (toG fdecls') fdecls'), P $ zip [1..] fdecls')
      uniq :: a -> State (Int,[(a,Int)]) Int
      uniq v = do { (i,bs) <- get; put (i+1, (v,i):bs); return i }
      uniq' :: State (Int, [a]) Int
      uniq'  = do { (i,bs) <- get; put (i+1, bs); return i }
      nameOf v = 
          do { (i,bs) <- get;
             ; case lookup v bs of 
                 Just j -> return j
                 Nothing ->
                     do k <- uniq v 
                        return k }
      h [] td = return ([],td)
      h (CST.FDecl f pats e:fs) td = 
          do (pats',td') <- hps pats td
             e'          <- he e (concatMap binds pats')
             (fs',td'')  <- h fs td'
             return ((f,pats',e'):fs',td'')
      hps [] td = return ([],td)
      hps (p:ps) td = 
          do (p',  td')  <- hp p td
             (ps', td'') <- hps ps td'
             return (p':ps', td'')
      hp (CST.PCon c r p) td =
          do (r', td')  <- hpr r td
             (p', td'') <- hp p td'
             i <- uniq'
             j <- uniq'
             return $ (PCon i c (PRec j r') p', td'')
      hp (CST.PApp p1 p2) td =
          do (p1', td')  <- hp p1 td
             (p2', td'') <- hp p2 td'
             i <- uniq'
             return $ (PApp i p1' p2', td'')
      hp (CST.PVar v t) td =
          do { (t',td') <- g t td
             ; case t' of
                 DS i ->
                     return $ (PVar i v, td')
                 _ ->
                     do i <- uniq'
                        return $ (PVar i v, (i,t'):td')
             }
      hp (CST.PLit []) td = hp (CST.PNil) td 
      hp (CST.PLit s) td =
          do cs <- mapM (\c -> do { i <- uniq'; return $ (PChar i c) }) s 
             r <- foldM1 (\a b-> do { i <- uniq'; return $ (PApp i a b) }) cs
             return (r, td)
      hp (CST.PNil) td = 
          do { i <- uniq'; return (PNil i, td) }
      hpr (CST.RCons (k,p) r) td = 
           do (p',td')  <- hp p td
              (r',td'') <- hpr r td'
              return $ ((k,p'):r', td'')
      hpr (CST.REmp) td = return $ ([], td)
--       hp (CST.PCon c r p) td = 
--           do (r',td')  <- hpr r td
--              (p',td'') <- hp p td'
--              return $ (CST.PCon c r' p', td'')
--       hp (CST.PApp p1 p2) td =
--           do (p1',td')  <- hp p1 td
--              (p2',td'') <- hp p2 td'
--              return $ (CST.PApp p1' p2', td'')
--       hp (CST.PVar v t) td = 
--           do { (t',td') <- g t td
--              ; case t' of 
--                  DS i ->
--                      return $ (CST.PVar v (CST.TFun i), td')
--                  _ ->
--                      do  i <- uniq'
--                          return $ (CST.PVar v (CST.TFun i), (i,t'):td')
--              }
--       hp (CST.PLit s) td = 
--           do return $ (CST.PLit s, td)
--       hp (CST.PNil) td = return $ (CST.PNil,td)
--       hpr (CST.RCons (k,p) r) td = 
--           do (p',td')  <- hp p td
--              (r',td'') <- hpr r td'
--              return $ (CST.RCons (k,p') r', td'')
--       hpr (CST.REmp) td = return $ (CST.REmp, td)
--      he :: CST.Exp String -> [(String,Int)] -> State (Int, [(String,Int)]) (IdExp Int String)
      he x bind = he' x 
          where 
            he' (CST.ECon b r e) = 
                do i  <- uniq'
                   r' <- he' r
                   e' <- he' e
                   return $ (ECon i b r' e')
            he' (CST.EFun f vs) = 
                do i <- uniq'
                   return $ (EFun i f vs)
            he' (CST.EApp (CST.EApp e1 e2) e3) = he' (CST.EApp e1 (CST.EApp e2 e3)) -- FIXME
            he' (CST.EApp CST.ENil e) = he' e
            he' (CST.EApp e CST.ENil) = he' e 
            he' (CST.EApp e1 e2) =
                do i <- uniq'
                   e1' <- he' e1 
                   e2' <- he' e2
                   return $ (EApp i e1' e2')
            he' (CST.EVar v) =
                case lookup v bind of 
                  Just i -> return $  EVar i v
                  _      -> undefined
            he' (CST.ELit []) = do { i <- uniq'; return $ EEps i }
            he' (CST.ELit s) = 
                do es <- mapM (\a -> do { i <- uniq'; return $ EChar i a}) s
                   foldM1 (\a b -> do { i <- uniq';return $ EApp i a b}) es
            he' (CST.ENil) = do { i <- uniq'; return $ EEps i }
            he' (CST.ERec rs) = 
                do es <- mapM he' (map snd rs)
                   i  <- uniq'
                   return $ ERec i (zip (map fst rs) es)
      f [] td = return ([],td)
      f (CST.TDecl t te:ts) td = 
          do t'  <- nameOf t
             (te',td')  <- g te td
             (ts',td'') <- f ts td'
             return $ ( (t',te'):ts', td'')
      g (CST.TApp t1 (CST.TNil)) td = g t1 td
      g (CST.TApp (CST.TNil) t1) td = g t1 td
      g (CST.TApp (CST.TApp t1 t2) t3) td = g (CST.TApp t1 (CST.TApp t2 t3)) td 
      g (CST.TCon c r t) td = 
          do (r',td1) <- gr r td
             (t',td2) <- g t  td1
             return $ (DT c r' t', td2)
      g (CST.TAnyLabel r t) td =
          do (r',td1) <- gr r td
             (t',td2) <- g t td1
             return $ (DAT r' t', td2)
      g (CST.TApp t1 t2) td = 
          do (t1',td1) <- g t1 td
             (t2',td2) <- g t2 td1
             return $ (DA t1' t2', td2)
      g (CST.TSel t1 t2) td = 
          do (t1',td1) <- g t1 td
             (t2',td2) <- g t2 td1
             i <- uniq'
             return $ (DS i, (i,t1'):(i,t2'):td2)
      g (CST.TFun a) td = 
          do i <- nameOf a
             return $ (DS i, td)
      g (CST.TRec r) td = gr r td
      g (CST.TChar c) td =
          return $ (DC c, td)
      g (CST.TAnyChar) td = return $ (DAnyC, td)
      g (CST.TStar t) td = 
          do i <- uniq' 
             (t',td') <- g t td
             return $ (DS i, (i,(DA t' (DS i))):(i,DE):td')
      g (CST.TNil) td = return $ (DE,td)
      gr (CST.TRFun a) td = 
          do i <- nameOf a
             return $ (DS i, td)
      gr x td = 
          do (rs, ng, td') <- gr' x td []
             return $ (DR (sort rs) ng, td')
          where
            gr' (CST.TREmp) td _  = return ([], DNothing, td)
            gr' (CST.TRRest) td ng = return ([], DWithout ng, td)
            gr' (CST.TRCons (k,t) r) td ng = 
                do (t',td') <- g t td
                   (rs', ng', td'') <- gr' r td' (k:ng)
                   return $ ((k,t'):rs', ng', td'')