module Automaton where

import Data.List
import Control.Monad.State
import Debug.Trace
import Data.Maybe
import Value
import Util

class WEq a where
    compatible :: a -> a -> Bool

newtype Automaton a = Automaton { transitions:: [(a,TypeExp a)] }


instance Functor Automaton where
    fmap f (Automaton transes) = Automaton $ map (\(x,tr) -> (f x, fmap f tr)) transes 

instance Show a => Show (Automaton a) where
    show (Automaton a) = showRed a

showRed xys = foldr (\a r->a++"\n"++r) [] $ zipWith (\x y -> f x ++ " ==> " ++ y) ls rs 
    where ls = map (show.fst) xys
          rs = map (show.snd) xys
          ml = maximum $ map length ls
          f s = s ++ replicate (ml - length s) ' '
          

data TypeExp a 
    = TT (Node) [a]
    | TN a 
    | TR [(String,a)] TS -- ^FIXME: Under this formalization, we hardly treate records with infinite keys.
    | TEps
      deriving (Eq,Ord)


instance Show a => Show (TypeExp a) where
    show (TT n qs)  = show n ++ show qs
    show (TN a)     = show a
    show (TR rs ng) = "{"++implode ":" (map (\(x,y) -> x++": "++show y) rs)++";"++show ng++"}"
    show TEps       = "$"

data TS = RNothing
        | RWithout [String]
          deriving (Show, Eq, Ord)

data Node 
    = NLabel String -- ^Trinary
    | FLabel String -- ^Label with Fixed Arity (Should be Implement)
    | NChar  Char   -- ^Charactor
    | AnyLabel      -- ^AnyLabel for NLabel
    | AnyChar       -- ^AnyChar
    | NSep1         -- ^Separator 1 (Internaly Used)
    | NSep2         -- ^Separator 2 (Internaly Used)
    | NNSep          -- ^Non Separator (Internaly Used)
      deriving (Ord,Eq)

instance WEq Node where
    compatible (NLabel l) (NLabel l') = l == l'
    compatible (AnyLabel) (NLabel _)  = True
    compatible (NLabel _) (AnyLabel)  = True
    compatible (AnyLabel) (AnyLabel)  = True
    compatible (NChar c)  (NChar c')  = c == c'
    compatible (AnyChar)  (NChar _)   = True
    compatible (NChar _)  (AnyChar)   = True
    compatible (AnyChar)  (AnyChar)   = True
    compatible NSep1 NSep1 = True
    compatible NSep2 NSep2 = True
    compatible NNSep NSep1 = False
    compatible NNSep NSep2 = False
    compatible NSep1 NNSep = False
    compatible NSep2 NNSep = False
    compatible NNSep _     = True
    compatible _ NNSep     = True
    compatible _ _         = False

unify (NLabel l) (AnyLabel)  = NLabel l
unify (AnyLabel) (NLabel l)  = NLabel l
unify (AnyLabel) (AnyLabel)  = AnyLabel
unify (NLabel l) (NLabel l') = NLabel l 
unify (AnyChar)  (NChar c)   = NChar c
unify (NChar c)  (AnyChar)   = NChar c
unify (NChar c)  (NChar c')  = NChar c
unify (AnyChar)  (AnyChar)   = AnyChar
unify (NNSep)    l           = l
unify l          (NNSep)     = l
unify l          l'          = l


instance Show Node where
    show (NLabel l) = "<"++show l++">"
    show (NChar c)  = '#':show c
    show (AnyLabel) = "<_>"
    show (AnyChar)  = "."
    show (NSep1)    = "/1/"
    show (NSep2)    = "/2/"
    show (NNSep)    = "!/_/"
--     show (TB c r t t') = "<"++show c++">"++show r++"["++show t++"]"++show t'
--     show (TC c t)    = '#':c:' ':show t
--     show (TAT t)     = '_':show t
--     show (TAC t)     = '.':show t
--     show (TEps)      = "$"
--     show (TN a)      = show a
--     show (TR rs ng)  = "{"++foldr (\(n,s) r -> n++":"++show s++","++r) [] rs++f ng++"}"
--         where f (RNothing) = ""
--               f (RWithout ng) = "!{"++show ng++ "}"
--     show (TSep1 a)   = "/1/ "++show a
--     show (TSep2 a)   = "/2/ "++show a
--     show (NSep a)    = "/_/ "++show a

instance Functor TypeExp where
    fmap f (TT n ts)     = TT n (map f ts)
    fmap f (TN n)        = TN $ f n
    fmap f (TR rs ng)    = TR (map (\(s,t) -> (s,f t)) rs) ng
    fmap f TEps          = TEps

-- instance Functor TypeExp where
--     fmap f (TB c r t t') = TB c (f r) (f t) (f t')
--     fmap f (TC c t)      = TC c (f t)
--     fmap f (TAT t)       = TAT (f t)
--     fmap f (TAC t)       = TAC (f t)
--     fmap f TEps          = TEps
--     fmap f (TN a)        = TN $ f a
--     fmap f (TR rs ng)    = TR (map (\(s,t) -> (s,f t)) rs) ng


states :: (Eq a, Ord a) => Automaton a -> [a]
states (Automaton am) = states' am
states' am = snub $ concatMap f am
    where f (x,TT _ ls) = x:ls
          f (x,TN y)    = [x,y]
          f (x,TEps)    = [x]
          f (x,TR rs ng)      = x:map (\(x,y) -> y) rs

snub :: (Ord a, Eq a) => [a] -> [a]
snub ls = f (sort ls) []
    where
      f [] z = []
      f [x] z = x:z
      f (x:y:xs) z = 
          if x == y then 
              f (y:xs) z
          else
              f (y:xs) (x : z)

{-|
 Reduce Machine
-}
reduceMachine (Automaton transes) =
    fx [] [] transes 
    where
      fx redState redTrans curTrans =
          let (redState',redTrans',restTrans) = next redState redTrans curTrans []
          in if length redState == length redState' then
                 Automaton redTrans'
             else
                 fx redState' redTrans' restTrans 
      next redState redTrans [] rest = (redState, redTrans, rest)
      next redState redTrans (tr:trs) rest = 
          case tr of 
            (x,TEps) ->
                next (x:redState) (tr:redTrans) trs rest
            (x,TT l qs) ->
                if all (\q -> q `elem` redState) qs then
                    next (x:redState) (tr:redTrans) trs rest
                else
                    next redState redTrans trs (tr:rest)
            (x,TR rs ng) ->
                if all (\q -> q `elem` redState) (map snd rs) then
                    next (x:redState) (tr:redTrans) trs rest
                else
                    next redState redTrans trs (tr:rest)
                    

          

 
{-|
 Epsilon Elimination
-}
eliminate_epsrules (Automaton transes) = 
    Automaton $ snub $ f eps_closures
    where
      eps_rules   = filter (\(x,y) -> case y of { TN _ -> True; _ -> False }) transes
      neps_rules  = filter (\(x,y) -> case y of { TN _ -> False; _ -> True }) transes
      all_states  = states' transes
      eps_closures = [ (q, eps_closure [q] eps_rules) | q <- all_states ]
      eps_closure qs trs = 
          let (trs1, trs2) = partition (\(x,y) -> x `elem` qs) trs
              qs' = map (\(x,TN n) -> n) trs1
          in case trs1 of 
               [] -> snub qs
               _  -> eps_closure (qs'++qs) trs2
      f [] = []
      f ((q,qs):closures) = 
          let trs = filter (\(x,y) -> x `elem` qs) neps_rules
          in (map (\(x,y) -> (q,y)) trs) ++ f closures

checkEmpty q (Automaton transes_) = 
    not $ evalState (ck q) ([],[])
       where
         transes = accessibleRules [q] transes_
         ck q = do { (proced,nonemp) <- get
                   ; if q `elem` proced then
                         return False
                     else if q `elem` nonemp then
                         return True
                     else 
                         do put (q:proced, nonemp)
                            b <- orCheck checkee
                            (_, nonemp') <- get
                            put (proced,nonemp')
                            return b
                   }
             where
               qtranses = filter (\(x,t) -> x == q) transes
               checkee = [ snub $ sts tr | (_,tr) <- qtranses ]
               sts (TT _ qs) = qs
               sts (TN q)    = [q]
               sts (TEps)    = []
               sts (TR rs _) = map snd rs
               orCheck [] = return False
               orCheck (qs:checkee) = 
                   andCheck qs >>= (\b ->
                                        if b then 
                                            return True
                                        else
                                            orCheck checkee)
               andCheck [] = return $ True
               andCheck (q:qs) = 
                   do { 
                   ; b <- ck q 
                   ; if b then 
                         do (proced, nonemp) <- get 
                            put (proced, q:nonemp)
                            andCheck qs
                     else
                         return False
                   }

{-|
 Checking Vertical Overlap for given two states.

 Let q,q' are given update state, 
 calculate whether or not (q,q') is accesible in the self product automaton.

 The function returns True    if (q,q') is accesible.
              returns False   otherwise

 The function is defined by State Monad with accelaration reason.
 The state is a pair of lists of a pair of states.
 The first component of the pair represents states that are assumed not to be overlapped.
 The second component of the pair represents states that are assumed to be overlaped.

 * TODO: This function is not well tested yet.

 * TODO: Records with infinite keys.

-}
checkOverlap q q' (Automaton transes) = 
    ck q q'
       where
         transes1 = accessibleRules [q] transes
         transes2 = accessibleRules [q'] transes
         ck q q' = do { (proced, overlaps) <- get
                      ; if (q,q') `elem` proced then
                            return False
                        else if (q,q') `elem` overlaps then
                            return True
                        else
                            do put ((q,q'):proced, overlaps)
                               b <- orCheck checkee
                               (_,overlaps') <- get
                               put (proced, overlaps')
                               return b
                      }
             where
               qtranses  = filter (\(x,t) -> x==q ) transes1
               qtranses' = filter (\(x,t) -> x==q') transes2
               checkee = [ snub $ sts tr tr' | (_,tr) <- qtranses, (_,tr') <- qtranses',  compatible tr tr' ]
               sts (TR rs RNothing) (TR rs' RNothing) = zip (map snd rs) (map snd rs')
               sts (TR rs RNothing) (TR rs' (RWithout vs')) = 
                   concatMap (\(k,q) -> case lookup k rs' of
                                          Just q' -> [(q,q')]
                                          Nothing -> []) rs
               sts (TR rs (RWithout vs)) (TR rs' RNothing) = 
                   concatMap (\(k,q') -> case lookup k rs of
                                          Just q -> [(q,q')]
                                          Nothing -> []) rs'
               sts (TR rs (RWithout vs)) (TR rs' (RWithout vs')) = 
                   concatMap (\(k,q) -> case lookup k rs' of
                                          Just q' -> [(q,q')]
                                          Nothing -> []) rs
                                 ++ concatMap (\(k,q') -> case lookup k rs of
                                                            Just q -> [(q,q')]
                                                            Nothing -> []) rs'
               sts (TT _ qs) (TT _ qs')         = zip qs qs'
               sts TEps TEps = []
               sts _ _ = []
               orCheck [] = return False
               orCheck (qs:checkee) = 
                   andCheck qs >>= (\b ->
                                        if b then 
                                            return True
                                        else
                                            orCheck checkee)
               andCheck [] = return $ True
               andCheck ((q,q'):qs) = 
                   do { 
                   ; b <- ck q q' 
                   ; if b then 
                         do (proced, overlaps) <- get 
                            put (proced, (q,q'):overlaps)
                            andCheck qs
                     else
                         return False
                   }

checkOverlap3 q1 q2 q3 (Automaton transes) = 
    ck q1 q2 q3
       where
         transes1 = accessibleRules [q1] transes
         transes2 = accessibleRules [q2] transes
         transes3 = accessibleRules [q3] transes
         ck q1 q2 q3 = 
             do { (proced, overlaps) <- get
                ; if (q1,q2,q3) `elem` proced then
                      return False
                  else if (q1,q2,q3) `elem` overlaps then
                      return True
                  else
                      do  put ((q1,q2,q3):proced, overlaps)
                          b <- orCheck checkee
                          (_,overlaps') <- get
                          put (proced, overlaps')
                          return $ b
                }
             where
               qtranses1 = filter (\(x,t) -> x==q1) transes1
               qtranses2 = filter (\(x,t) -> x==q2) transes2
               qtranses3 = filter (\(x,t) -> x==q3) transes3
               checkee = [ snub $ sts tr1 tr2 tr3 | (_,tr1) <- qtranses1, (_,tr2) <- qtranses2, (_,tr3) <- qtranses3,  compatible tr1 tr2, compatible tr2 tr3 ]
               sts (TT _ qs1) (TT _ qs2) (TT _ qs3) = zip3 qs1 qs2 qs3
               sts (TR rs1 RNothing) (TR rs2 RNothing) (TR rs3 RNothing) = zip3 (map snd rs1) (map snd rs2) (map snd rs3)
               sts (TR rs1 _) (TR rs2 _) (TR rs3 _) = [ (q1,q2,q3) | (k1,q1) <- rs1, (k2,q2) <- rs2, k1==k2, (k3,q3) <- rs3, k2==k3 ]
--               sts (TR rs1 (RWithout vs1)) (TR rs2 (RWithout vs2)) (TR rs3 (RWithout vs3)) = undefined -- FIXME
               sts _ _ _ = []
               orCheck [] = return False
               orCheck (qs:checkee) = 
                   andCheck qs >>= (\b ->
                                        if b then 
                                            return True
                                        else
                                            orCheck checkee)
               andCheck [] = return $ True
               andCheck ((q1,q2,q3):qs) = 
                   do { 
                   ; b <- ck q1 q2 q3
                   ; if b then 
                         do (proced, overlaps) <- get 
                            put (proced, (q1,q2,q3):overlaps)
                            andCheck qs
                     else
                         return False
                   }

instance WEq (TypeExp a) where
    compatible (TT l qs) (TT l' qs') = length qs == length qs' && compatible l l'
    compatible (TR rs RNothing) (TR rs' RNothing) | length rs == length rs' = and $ zipWith (==) (sort$map fst rs) (sort$map fst rs')
    compatible (TR rs RNothing) (TR rs' (RWithout vs')) =
        all (\k -> k `elem` (map fst rs') || not (k `elem` vs')) (map fst rs) && all (\k -> k `elem` map fst rs) (map fst rs')
    compatible (TR rs (RWithout vs)) (TR rs' (RNothing)) =
        all (\k -> k `elem` (map fst rs) || not (k `elem` vs)) (map fst rs') && all (\k -> k `elem` map fst rs') (map fst rs)
    compatible (TR rs (RWithout vs)) (TR rs' (RWithout vs')) = 
        all (\k -> k `elem` (map fst rs') || not (k `elem` vs')) (map fst rs)
                && all (\k -> k `elem` (map fst rs) || not (k `elem` vs)) (map fst rs')            
    compatible TEps TEps = True
    compatible _ _ = False

{- |
  The function calculate the product of given two tree automata.
  NB: The returned automaton may not be reduced.
-}
-- productMachine :: (Ord a, Eq a) => Automaton a -> Automaton a -> Automaton (a,a)
productMachine (Automaton transes) (Automaton transes') = -- trace ("PRODUCT:\n"++show (Automaton transes) ++ ","++show (Automaton  transes') ++ show (    Automaton $ snub $ fst $ (execState cs) ([],[]))) $ 
    reduceMachine $ Automaton $ 
       (transitions $ fmap (\q->(Left q,Left q)) (Automaton transes))
         ++(transitions $ fmap  (\q->(Right q,Right q)) (Automaton transes'))++(snub $ fst $ (execState cs) ([],[]))
    where
      cs = mapM_ contr $ [ (Left q,  Right q') | q <- states' transes, q' <- states' transes' ]
      contr (q,q') = 
          do { constructsRule (q,q') checkee }
          where
            woTag y = case y of { Left x -> x; Right x -> x } 
            qtranses  = filter (\(x,t) -> x==woTag q) transes
            qtranses' = filter (\(x,t) -> x==woTag q') transes'
            checkee = [ st tr tr' | (_,tr) <- qtranses, (_,tr') <- qtranses', compatible tr tr' ]
            st (TT l qs) (TT l' qs') = TT (unify l l') $ zipWith (\q q' -> (Left q, Right q'))  qs qs'
            st (TR rs RNothing) (TR rs' RNothing) = TR (zipWith (\(k,q) (_,q') -> (k,(Left q, Right q'))) rs rs') RNothing
            st (TR rs RNothing) (TR rs' (RWithout [])) =
                TR (map (\(k,q) -> (k,(Left q,Left q))) rs) RNothing
            st (TR rs (RWithout [])) (TR rs' (RNothing)) =
                TR (map (\(k,q) -> (k,(Right q,Right q))) rs') RNothing
            st (TR rs (RWithout [])) (TR rs' (RWithout [])) = 
                TR (zipWith (\(k,q) (_,q') -> (k,(Left q, Right q'))) rs  rs') (RWithout [])
            st (TR rs (RWithout vs)) (TR rs' (RWithout vs')) = undefined -- FIXME
            st TEps TEps = TEps
            st t1 t2 = trace (show t1 ++ show t2) $ error "A"
            appendTrans (q,tr) = 
                do (transes, proced) <- get
                   put ((q,tr):transes, proced)
            check q = do (transes, proced) <- get
                         put (transes, q:proced)
            constructsRule q [] = return ()
            constructsRule q (trs:checkee) = 
                do appendTrans (q,trs)
                   constructsRule q checkee

    
                        
data Tag a
    = L1 a
    | L2 a
    | R1 a 
    | R2 a 
    | O1 a
    | O2 a
    | O3 a
    | N3 Int
    | D  a
      deriving (Eq,Ord,Show)

-- Assumption: the input is already reduced, determinized.
checkSingletonD qs_ (Automaton transes) = -- (\s -> trace (show (qs,Automaton transes)++" ==> " ++ show s) s) $ 
    if length qs /= 1 then
        Nothing 
    else 
        let [q] = qs in allSingle q []
    where
      qs = intersect qs_ (states $ Automaton transes) 
      allSingle q proced 
          | q `elem` proced = Nothing
          | otherwise =
              case [ tr | (x,tr) <- transes, x == q ] of
                [tr] ->
                    do { ts' <- mapM (\p -> allSingle p (q:proced)) (sts tr)
                       ; return $ mkTree tr ts' }
                _ ->
                    Nothing 
          where
            sts (TT AnyChar _)   = [q]
            sts (TT AnyLabel _)  = [q]
            sts (TT _ qs)        = qs
            sts (TEps)           = []
            sts (TR rs RNothing) = map snd rs
            sts (TR _ _)         = [q]
            mkTree (TT l _) ts   = mkT l ts
            mkTree (TR rs _) ts  = mkR rs ts
            mkTree (TEps) ts     = VEps
            mkT (NChar c ) [t]        = VApp (VChar c) t
            mkT (NLabel s) [t1,t2,t3] = VApp (VCon s t1 t2) t3
            mkR rs ts = VRec $ zipWith (\(k,q) t-> (k,t)) rs ts

                         


checkSingleton :: (Eq a, Ord a) => a -> Automaton a -> Maybe Val
checkSingleton qinit (Automaton transes') =
    fix [] []
    where
      transes = accessibleRules [qinit] transes'
      fix counts trees =
          let (counts',trees',changed, flag) = fixStep counts trees 
          in if flag then
                 Nothing
             else
                 if changed then
                     fix counts' trees'
                 else
                     case lookup qinit counts of
                       Just 1 -> lookup qinit trees
                       _      -> Nothing
      fixStep counts trees =
          foldr (\(x,tr) (cs,ts,cd,flg) ->
                     if flg then 
                         (cs,ts,cd,flg)
                     else
                         case lookup x counts of 
                           Just 2 -> (cs,ts,cd,flg)
                           _ -> g (x,tr) cs ts cd) (counts,trees,False,False) transes
      g (x,tr) counts trees changed = f (x,tr)
          where
            f (x,TT AnyLabel qs)
                | all zero qs = (counts,trees,changed,False)
                | x == qinit  = ((qinit,2):counts,trees,False,True)
                | x /= qinit  = ((x,2):counts,trees,True,False)
            f (x,TT AnyChar qs) = f (x, TT AnyLabel qs) 
            f (x,TT l qs) 
                | all mt0 qs && any two qs && x == qinit = ((qinit,2):counts,trees,False,True)
                | all mt0 qs && any two qs = ((x,2):counts,trees,True,False)
                | all one qs = case lookup x counts of
                                 Just 1 -> 
                                     let t  = mkT l qs
                                         t' = lookup' x trees 
                                     in if t /= t' then 
                                            ((x,2):counts,trees,True,False)
                                        else
                                            (counts,trees,False,False)
                                 _ ->
                                     ((x,1):counts,(x,mkT l qs):trees,True,False)
            f (x,TR rs (RWithout _)) 
                | all zero (map snd rs) = (counts,trees,changed,False)
                | x == qinit  = ((qinit,2):counts,trees,False,True)
                | x /= qinit  = ((x,2):counts,trees,True,False)
            f (x,TR rs RNothing) 
                | all mt0 (map snd rs) && any two (map snd rs) && x == qinit =  ((qinit,2):counts,trees,False,True)
                | all mt0 (map snd rs) && any two (map snd rs) = ((x,2):counts,trees,True,False)
                | all one (map snd rs) = case lookup x counts of 
                                 Just 1 ->
                                     let r  = mkR rs 
                                         r' = lookup' x trees
                                     in if r /= r' then
                                            ((x,2):counts,trees,True,False)
                                        else
                                            (counts,trees,False,False)
                                 _ ->
                                     ((x,1):counts,(x,mkR rs):trees, True, False)
            f (x,TEps) = case lookup x counts of 
                           Just 1 -> 
                               if VEps /= lookup' x trees then
                                   ((x,2):counts,trees,True,False)
                               else 
                                   (counts,trees,False,False)
                           _ ->
                               ((x,1):counts,(x,VEps):trees, True, False)
            mt0 x = case lookup x counts of { Just _ -> True; Nothing -> False }
            one x = case lookup x counts of { Just 1 -> True; _ -> False}
            two x = case lookup x counts of { Just 2 -> True; _ -> False}
            zero x = case lookup x counts of { Nothing -> True ; _ -> False}
            mkT (NChar c ) [q]        = VApp (VChar c) (lookup' q trees)
            mkT (NLabel s) [q1,q2,q3] = VApp (VCon s (lookup' q1 trees) (lookup' q2 trees)) (lookup' q3 trees)
            mkR rs = VRec $ map (\(k,q) -> (k,lookup' q trees)) rs


accessibleRules qs transes = 
    snub $ f qs transes
    where
      f qs drs =  -- trace (show (states' $ fst $ partition (\(x,tr) -> x `elem` qs) drs)) $ 
          let (drs1, drs2) = partition (\(x,tr) -> x `elem` qs) drs
          in case drs2 of
               [] -> drs1
               _  ->
                   case drs1 of
                     [] -> []
                     _  -> drs1++f (states' drs1++qs) drs2
hFirstNodes q (Automaton transes) =
  snub $ concatMap (\(x,tr)->
                        if x == q then 
                            case tr of
                              TT l qs  -> [l]
                              TR rs ng -> [AnyLabel, AnyChar] -- FIXME
                              TEps     -> []
                        else
                            []) transes
    
hAcceptableNodes q (Automaton transes) = 
    snub $ evalState (f q) []
    where
      f q = 
          do { proced <- get 
             ; if q `elem` proced then
                   return []
               else
                   do let (qs,next) = g $ map snd $ filter (\(q',tr)->q==q') transes
                      put (q:proced)
                      qss' <- mapM f next                      
                      return $ qs ++ concat qss'
             }
          where 
            g (TT l qs:xs)       = let (s,t) = g xs in (l:s,last qs:t)
            g (TR rs ng:xs)      = ([AnyLabel, AnyChar],[]) -- Records Are Not Supported Well
            g (TEps:xs)          = g xs
            g []                 = ([],[])

hSingle q (Automaton transes) = 
    all goEps next
    where
      next = [ nx (TT l qs) | (x,TT l qs) <- transes, x == q ]
      nx (TT (NLabel _) [_,_,q]) = q
      nx (TT (AnyLabel) [_,_,q]) = q
      nx (TT (NChar _)  [q])     = q
      nx (TT (AnyChar)  [q])     = q
      goEps q = null [ () | (x,tr) <- transes, x == q, tr /= TEps ]

-- g ((TB c _ _ q2):xs) = let (s,t) = g xs in (Tree c:s,q2:t)
--                 g ((TAT q2):xs)      = let (s,t) = g xs in (AllTree:s,q2:t)
--                 g ((TC c q2):xs)     = let (s,t) = g xs in (Char c:s,q2:t)
--                 g ((TAC q2):xs)      = let (s,t) = g xs in (AllChar:s,q2:t)
--                 g (TR rs ng:xs)      = ([AllTree, AllChar],[]) -- Records Are Not Supported Well
--                 g (TEps:xs)          = g xs
--                 g []                 = ([],[])

        
{- |
 
-}
lightWeightOverlapCheck q q' (Automaton transes) = -- trace (show q++" "++show q'++":"++show may1 ++ ", " ++ show first2) $ 
    (not single) && any (`elem` may1) first2 && olTestS may1 may2
        where
          single = hSingle q (Automaton transes)
          may1   = hAcceptableNodes q  (Automaton transes)
          first2 = hFirstNodes q' (Automaton transes)
          may2   = hAcceptableNodes q' (Automaton transes)
          olTestS ns ns' = not $ null $ intersect ns ns'        

checkOverlapH q q' (Automaton transes_) = 
    if lightWeightOverlapCheck q q' (Automaton transes) then -- trace (show $ Automaton $ accessibleRules [((L1 q, O2 q))] $ transitions $ productMachine (Automaton m1) (Automaton m2)) $ 
        checkOverlap3 (L1 q) (O2 q) (O3 q) (Automaton $ m1++m2++t)
--        checkOverlap (L1 q, O2 q) (O3 q, O3 q) (Automaton $ t ++ m)
    else
        return False
    where 
      transes = accessibleRules [q,q'] $ transitions $ (Automaton transes_)
      am1 = snub $ accessibleRules [L1 q] $ (concatMap (g L1) transes ++ concatMap (g R1) transes ++ [ (L1 q, TT NSep1 [R1 q]) | q <- states' transes])
      am2 = snub $ accessibleRules [L2 q'] $ (concatMap (g L2) transes ++ concatMap (g R2) transes ++ [ (L2 q, TT NSep2 [R2 q]) | q <- states' transes]) --  FIXME: redundunt: R2 & D --> R2
      g k (q,tr) =
          case tr of
            TT (NLabel l) [q1,q2,q3] ->
                [(D q, TT (NLabel l) [D q1, D q2, D q3]), (k q, TT (NLabel l) [D q1, D q2, k q3])]
            TT AnyLabel [q1,q2,q3] ->
                [(D q, TT AnyLabel [D q1, D q2, D q3]), (k q, TT AnyLabel [D q1, D q2, k q3])]
            TT (NChar c) [q1] ->
                [(D q, TT (NChar c) [D q1]), (k q, TT (NChar c) [k q1])]
            TT AnyChar [q1] ->
                [(D q, TT AnyChar [D q1]), (k q, TT AnyChar [k q1])]
            TN q1 ->
                [(D q, TN (D q1)), (k q, TN (k q1))]
            TR rs ng -> 
                [(D q, TR (map (\(k,t) -> (k,D t)) rs) ng)]
            TEps -> 
                [(D q,TEps), (k q, TEps)]
      Automaton m1' = fmap O1 (Automaton $ accessibleRules [q' ] transes)
      Automaton m2' = Automaton (concatMap (g O2) (accessibleRules [q] transes))
      m1  = (concatMap f am1) ++ m1'
          where f (R1 p, TEps) = [(R1 p, TT (NSep2) [p']) | p' <- states' m1', p' == O1 q']
                f t            = [t]
      m2  = (concatMap f m2')++am2
          where f (O2 p,TEps) = [(O2 p, TT (NSep1) [L2 p']) | (L2 p') <- states' am2, p' == q']
                f t        = [t]
      n i = (N3 i, N3 i)
      t = (transitions $ fmap O3 (Automaton transes))
          ++ concat [ [(O3 q, TT NSep1 [N3 0]), (N3 0, TT AnyChar [N3 1]), (N3 0, TT AnyLabel [N3 4, N3 5, N3 1]),
                       (N3 1, TT AnyChar [N3 1]), (N3 1, TT AnyLabel [N3 4, N3 5, N3 1]),  (N3 1, TT NSep2 [O3 q'])] | (q,TEps) <- transes ]
          ++ [ (N3 4, TR [] (RWithout [])), (N3 5, TT AnyChar [N3 5]), (N3 5, TT AnyLabel [N3 4, N3 5, N3 5]), (N3 5, TEps)]
--       t = (transitions $ fmap (\x -> (O3 x, O3 x)) (Automaton transes))
--             ++ concat [ [((O3 q, O3 q), TT NSep1 [(n 0)]), (n 0, TT NNSep [(n 1)]), 
--                          (n 1, TT NNSep [(n 1)]), (n 1, TT NSep2 [(O3 q', O3 q')])]  | (q,TEps) <- transes ]
--      m = accessibleRules [((L1 q, O2 q))] $ transitions $ productMachine (Automaton m1) (Automaton m2)


-- checkEquiv :: (Ord a, Ord b, Eq a, Eq b) => [a] -> [b] -> Automaton a -> Automaton b -> Bool
checkEquiv qs qs' am am' = -- trace ("\n\n\n\nEQ:\n"++show (qs1,am1) ++ show (qs2, am2) ++ " -----> " ++ show (ans) ) $ 
    checkEquivD qs1 qs2 am1 am2
    where 
      (qs1,am1) = uncurry minimize $ uncurry numerizeState $ determinize qs  am
      (qs2,am2) = uncurry minimize $ uncurry numerizeState $ determinize qs' am'

checkEquivD qs1 qs2 am1 am2 =
    length qs1 == length qs2 && length (transitions am1) == length (transitions am2) && checkInclude qs1 qs2 am1 am2 && checkInclude qs2 qs1 am2 am1

checkInclude qs1 qs2 (Automaton transes1) (Automaton transes2) = 
    evalState (ck qs1 qs2) ([],[])
    where
      ck qs1 qs2 = 
          foldM (\b q1 -> 
                     if b then
                         foldM (\b q2 ->
                                    if b then
                                        return True
                                    else
                                        chk q1 q2) False qs2
                     else 
                         return False) True qs1
      chk q1 q2 =
          do { (proced,inc) <- get
             ; if (q1,q2) `elem` proced then
                   return True
               else if (q1,q2) `elem` inc then
                   return True
               else -- trace (show [ (x1,tr1,[ snub $ sts tr1 tr2 | (x2,tr2) <- transes2, x2==q2, compatible tr1 tr2 ]) | (x1,tr1) <- transes1, x1==q1 ]) $ 
                   do put ((q1,q2):proced,inc)
                      b <- and1Check checkee
                      (_,inc') <- get
                      put (proced, (if b then ((q1,q2):) else id) inc')
                      return b 
             }
          where
--            cond    = null [ () | (_,tr1) <- [(x1,tr1) <- transes1, x1==q1], (x2,tr2) <- transes2, x2==q2, not (compatible tr1 tr2) ]
            checkee = [ [ snub $ sts tr1 tr2 | (x2,tr2) <- transes2, x2==q2, compatible tr1 tr2 ] | (x1,tr1) <- transes1, x1==q1 ]
            sts (TR rs RNothing) (TR rs' RNothing) = zip (map snd rs) (map snd rs')
            sts (TR rs RNothing) (TR rs' (RWithout vs')) = 
                concatMap (\(k,q) -> case lookup k rs' of
                                       Just q' -> [(q,q')]
                                       Nothing -> []) rs
            sts (TR rs (RWithout vs)) (TR rs' RNothing) = 
                concatMap (\(k,q') -> case lookup k rs of
                                        Just q -> [(q,q')]
                                        Nothing -> []) rs'
            sts (TR rs (RWithout vs)) (TR rs' (RWithout vs')) = 
                concatMap (\(k,q) -> case lookup k rs' of
                                       Just q' -> [(q,q')]
                                       Nothing -> []) rs
                              ++ concatMap (\(k,q') -> case lookup k rs of
                                                         Just q -> [(q,q')]
                                                         Nothing -> []) rs'
            sts (TT _ qs) (TT _ qs')         = zip qs qs'
            sts TEps TEps = []
            sts _ _ = []
            and1Check [] = return True
            and1Check (qss:qsss) = 
                do { b <- orCheck qss
                   ; if b then
                         and1Check qsss
                     else
                         return False }
            andCheck [] = return True
            andCheck ((q1,q2):qs) = 
                do { b <- chk q1 q2
                   ; if b then
                         andCheck qs
                     else
                         return False }
            orCheck [] = return False
            orCheck (qs:qss) = 
                do { b <- andCheck qs
                   ; if b then
                         return True
                     else
                         orCheck qss } 
-- and [ or [ chk q1 q2 | q1 <- qs2 ] | q1 <- qs1 ]
      -- ck qs1 qs2 = trace ("EQS:" ++ show eqs) $ 
      --     and [ or [ (q1,q2) `elem` eqs | q2 <- qs2 ] | q1 <- qs1 ]
      --       where  eqs = equivFix [] 
      -- equivFix eqs =
      --     let eqs' = equivStep eqs 
      --     in if length eqs' == length eqs then
      --            eqs'
      --        else
      --            equivFix eqs'
      -- equivStep eqs = -- trace ("EQUIV:"++show eqs) $ 
      --     [ (x1,x2) | (x1,tr1) <- transes1, (x2,tr2) <- transes2, 
      --                 compatible tr1 tr2, all (`elem`eqs) (sts tr1 tr2) ] \\
      --     [ (x1,x2) | (x1,tr1) <- transes1, (x2,tr2) <- transes2,
      --                 almostCompatible tr1 tr2, not (compatible tr1 tr2), all (`elem`eqs) (sts tr1 tr2)]
      --     where
      --       almostCompatible (TT _ qs) (TT _ qs') = True
      --       almostCompatible (TR _ _) (TR _ _) = True
      --       almostCompatible (TEps)   (TEps)   = True
      --       almostCompatible _ _ = False
      --       sts (TR rs RNothing) (TR rs' RNothing) = zip (map snd rs) (map snd rs')
      --       sts (TR rs RNothing) (TR rs' (RWithout vs')) = 
      --           concatMap (\(k,q) -> case lookup k rs' of
      --                                  Just q' -> [(q,q')]
      --                                  Nothing -> []) rs
      --       sts (TR rs (RWithout vs)) (TR rs' RNothing) = 
      --           concatMap (\(k,q') -> case lookup k rs of
      --                                   Just q -> [(q,q')]
      --                                   Nothing -> []) rs'
      --       sts (TR rs (RWithout vs)) (TR rs' (RWithout vs')) = 
      --           concatMap (\(k,q) -> case lookup k rs' of
      --                                  Just q' -> [(q,q')]
      --                                  Nothing -> []) rs
      --                         ++ concatMap (\(k,q') -> case lookup k rs of
      --                                                    Just q -> [(q,q')]
      --                                                    Nothing -> []) rs'
      --       sts (TT _ qs) (TT _ qs')         = zip qs qs'
      --       sts TEps TEps = []
      --       sts _ _ = []


numerizeState qs am = (map f qs, am')
    where
      sorted_states = snub $ qs ++ states am 
      am' = fmap f am
      f n = elemIndex' n sorted_states 

minimize qs (Automaton transes) = -- (\s -> trace ("I:" ++ show (Automaton transes)++"O:"++show s) s) $ 
    let stgroup = divideFix [ r_states `intersect` qs, qs \\ r_states, r_states `intersect` (all_states \\ qs), (all_states \\ qs) \\ r_states  ]
        (qs',Automaton tr') = numerizeState (new_finals stgroup) (Automaton $ new_transes stgroup)
    in (qs' `intersect` (states' tr'),Automaton $ accessibleRules qs' tr')
       where
         all_states = states' transes
         r_states = map fst $ filter (\(x,tr) -> case tr of {TR _ _ -> True; _ -> False}) transes
         groupingBy eq []     = []
         groupingBy eq (x:xs) = let (ls,rs) = partition (eq x) xs in (x:ls):groupingBy eq rs
         isSameGroup stgroup q1 q2 = any (\v->q1`elem`v&&q2`elem`v) stgroup
         willBeSameGroup stgroup q1 q2 = 
             isSameGroup stgroup q1 q2  && all allMatched transQ1s && all allMatched' transQ2s
             where
               transQ1s = filter (\(x,tr)->case tr of {TT _ qs|q1`elem`qs->True;TR rs _|q1`elem`(map snd rs)->True;_->False}) transes
               transQ2s = filter (\(x,tr)->case tr of {TT _ qs|q2`elem`qs->True;TR rs _|q2`elem`(map snd rs)->True;_->False}) transes               
               sameExcept [] [] = True
               sameExcept (s:ss) (s':ss') | s==s'         = sameExcept ss ss'
                                          | s==q1&&s'==q2 = ss==ss'
                                          | otherwise     = False
               sameExcept _ _ = False
               allMatched (x,TEps)    = True -- and [ isSameGroup stgroup x x' | (x',TEps) <- transQ2s ]
               allMatched (x,TT l xs)  = 
                   let vs = [ x' | (x',TT l' xs') <- transQ2s, compatible l l', sameExcept xs xs']
                   in  not (null vs) && and [ isSameGroup stgroup x x' | x' <- vs ]                   
               allMatched (x,TR rs ng) = 
                   let vs = [ x' | (x',TR rs' ng') <- transQ2s, cmp rs ng rs' ng,  sameExcept (map snd rs) (map snd rs') ] 
                   in  not (null vs) && and [ isSameGroup stgroup x' x | x' <- vs ]
               allMatched' (x,TEps)    = True -- and [ isSameGroup stgroup x x' | (x',TEps) <- transQ2s ]
               allMatched' (x,TT l xs)  = 
                   let vs = [ x' | (x',TT l' xs') <- transQ1s, compatible l l', sameExcept xs xs']
                   in  not (null vs) && and [ isSameGroup stgroup x' x | x' <- vs ]                   
               allMatched' (x,TR rs ng) = 
                   let vs = [ x' | (x',TR rs' ng') <- transQ1s, cmp rs ng rs' ng,  sameExcept (map snd rs) (map snd rs') ] 
                   in  not (null vs) && and [ isSameGroup stgroup x' x | x' <- vs ]
               -- allMatched (x,TT l xs) = and [ not ( compatible l l' ) || not ( sameExcept xs xs' ) || 
               --                                isSameGroup stgroup x x' | (x',TT l' xs') <- transQ2s ]
               -- allMatched (x,TR rs ng) = and [ not ( cmp rs ng rs' ng' ) || not ( sameExcept (map snd rs) (map snd rs') ) ||
               --                                 isSameGroup stgroup x x' | (x',TR rs' ng') <- transQ2s ]
               cmp rs RNothing rs' RNothing = map fst rs == map fst rs'
               cmp rs (RWithout ng) rs' (RWithout ng') = not (null (ks `intersect` ks')) && null ((ks\\ks') `intersect`ng') && null ((ks'\\ks) `intersect`ng)
                   where ks  = map fst rs 
                         ks' = map fst rs'
               cmp _ _ _ _ = False
         divide stgroup = groupingBy (willBeSameGroup stgroup) all_states
         divideFix stgroup = -- trace (">>" ++ show stgroup) $ 
             let stgroup' = divide stgroup 
             in if length stgroup' == length stgroup then
                    stgroup'
                else
                    divideFix stgroup'
         new_transes stgroup = 
             snub $ map (\(x,tr) -> case tr of 
                                     TT l qs  -> (eqF stgroup x, TT l (map (eqF stgroup) qs))
                                     TR rs ng -> (eqF stgroup x, TR (map (\(k,q) -> (k,eqF stgroup q)) rs) ng)
                                     TEps     -> (eqF stgroup x, TEps)) transes
         new_finals stgroup =
             snub $ map (eqF stgroup) qs
         eqF [] q = []
         eqF (sts:x) q 
             | q `elem` sts = sts
             | otherwise    = eqF x q

{-
 This function determinize a given automaton.
 FIXME:
  * FIX attributes treatment
-}
determinize qs am = 
    (snub $ map (map (\i -> sstates !! i)) [ x | (x,tr) <- determinized_transes, any (`elem`x) (map (\q -> elemIndex' q sstates) qs) ],
     fmap (map (\i -> sstates !! i)) $ Automaton $ determinized_transes)
        where
          Automaton transes = eliminate_epsrules am
          sstates = snub (qs++states' transes)
          Automaton transes' = fmap (\n -> elemIndex' n sstates) (Automaton transes)
          toTR (keys,ng) qs = TR (zip keys qs) ng
          ttranses = filter (\(_,tr) -> case tr of {TT _ qs -> True; _ -> False }) transes'
          etranses = filter (\(_,tr) -> case tr of {TEps -> True; _ -> False}) transes'
          rtranses = filter (\(_,tr) -> case tr of {TR _ _ -> True; _ -> False}) transes'
          nsymbols = snub $ map (\(_,TT c qs) -> (c,length qs)) $ ttranses
          rsymbols = snub $ map (\(_,TR rs ng) -> (map fst rs, ng)) $ rtranses
          nextT c qs = 
              foldr (\(p,TT c' qs') (c'',r)->
                         if compatible c c' && and (zipWith elem qs' qs) then
                             (unify c c'', p:r)
                         else
                             (c'',r)) (c,[]) ttranses
          nextR (rs,n) qs = 
              let tq = concatMap (\(p,TR rq ng)->
                                      if rs == map fst rq && n == ng && and (zipWith elem (map snd rq) qs) then
                                          [p]
                                      else []) rtranses
              in snub tq
          determinize_step dstates dtranses = -- trace (show $ length dtranses) $ 
              let new_ttranses =
                      do (c,n) <- nsymbols
                         qs <- foldr (\_ r -> [(q:qs) | q <- dstates, qs <- r]) [[]] [1..n]
                         return (let (c',s) = nextT c qs
                                 in (s,TT c' qs))
                  new_rtranses =
                      do r <- rsymbols
                         qs <- foldr (\_ r -> [(q:qs) | q <- dstates, qs <-r]) [[]] [1..(length$fst r)]
                         return (let s = nextR r qs 
                                 in (s, toTR r qs))
                  dtranses' = filter (\(x,y) -> not (null x)) $ snub (new_ttranses ++ new_rtranses ++ [(s_etranses,TEps)])
              in if length dtranses' == length dtranses then
                     dtranses
                 else
                     determinize_step (states' dtranses') dtranses'    
          s_etranses = states' etranses
          determinized_transes = determinize_step ([s_etranses]) [(s_etranses,TEps)]
                             

renameTranses s (Automaton x) = Automaton $ renameTranses' s x

renameTranses' [] transes = transes
renameTranses' ((a,b):sts) transes =
    renameTranses' (map (\(x,y) -> 
                            if y == a then (x,b)
                            else           (x,y)) sts)
                   (map (\(x,t) -> (x,fmap 
                                        (\n -> if n == a then b
                                               else           n) t)) transes)

