module CST where

import Data.List
import Control.Monad.State
import Debug.Trace
import Data.Maybe
import Automaton
import Util

type Prog a = [Decl a]
 
data Decl a = TDecl String (TExp a)
            | FDecl String [Pat a String] (Exp a)
              deriving (Eq,Ord)

instance Show a => Show (Decl a) where
    show (TDecl t te) = "data "++show t++" = "++show te++"\n"
    show (FDecl f p e) = f++"("++show p++") = "++show e ++ "\n"


data Pat a b = PCon String (RPat a b) (Pat a b)
             | PApp (Pat a b) (Pat a b)
             | PVar b (TExp a)
             | PLit String
             | PNil 
             deriving (Eq,Ord)

instance (Show a, Show b) => Show (Pat a b) where
    show (PCon c r p) = "<"++c++">"++show r++"("++show p++")"
    show (PApp p1 p2) = show p1 ++ " " ++ show p2 
    show (PVar v t)   = show v ++ " :: " ++ show t 
    show (PLit s)     = show s
    show PNil         = "$"

data RPat a b = RCons (String,Pat a b) (RPat a b)
              | REmp 
              | RVar b -- FIXME
                deriving (Eq,Ord)

instance (Show a,Show b) => Show (RPat a b) where
    show REmp = ""
    show (RVar s) = show s 
    show xs = "{"++f xs++"}"
        where f (RCons (k,t) r) = show k ++ ":"++show t++ "," ++ show r
              f (REmp)          = ""


data NPat a b -- Patterns Associted with Typename
    = PC a String (NPat a b) (NPat a b) 
    | PA a a (NPat a b) (NPat a b)
    | PV a b
    | PR a [(String, NPat a b)]
    | PEps
      deriving (Eq,Ord)

data Exp a = ECon String (Exp a) (Exp a)
           | EFun String [String]
           | EApp (Exp a) (Exp a)
           | EVar String
           | ELit String 
           | ERec [(String,Exp a)] 
           | EUni (Exp a) (Exp a)         -- FIXME
           | ENil 
          deriving (Eq,Ord)

instance Show a => Show (Exp a) where
    show (ECon c r e) = "<"++c++">"++show r++"("++show e++")"
    show (EFun f vs)  = f ++ "(" ++ implode "," vs ++ ")"
    show (EApp e1 e2) = show e1 ++ " " ++ show e2
    show (EVar v)     = v
    show (ELit s)     = "\"" ++ s ++ "\""
    show (ERec r)     = "{" ++ implode "," (map (\(s,t)->s ++ ":" ++ show t) r) ++ "}"
    show (ENil)       = "$"
    show (EUni r1 r2) = show r1 ++ "#"++ show r2

data TExp a = TCon String (TRec a) (TExp a)
            | TApp (TExp a) (TExp a)
            | TSel (TExp a) (TExp a)
            | TFun a      
            | TRec (TRec a)
            | TChar Char  -- a string literal is a type.
--            | TAnyTree    -- any tree
            | TAnyLabel (TRec a) (TExp a)
            | TAnyChar    -- any char
            | TStar (TExp a)
            | TNil 
          deriving (Eq, Ord)

instance Functor (TExp) where
    fmap f (TCon c r e) = TCon c (fmap f r) (fmap f e)
    fmap f (TApp t1 t2) = TApp (fmap f t1) (fmap f t2)
    fmap f (TSel t1 t2) = TSel (fmap f t1) (fmap f t2)
    fmap f (TFun a)     = TFun $ f a 
    fmap f (TChar c)    = TChar c
    fmap f (TRec r)     = TRec $ fmap f r
--    fmap f TAnyTree     = TAnyTree
    fmap f (TAnyLabel r e) = TAnyLabel (fmap f r) (fmap f e)
    fmap f TAnyChar     = TAnyChar
    fmap f (TStar t)    = TStar (fmap f t)
    fmap f TNil         = TNil
         
instance Show a => Show (TExp a) where 
    show (TCon c r t) = "<"++c++">"++showR r++"("++show t++")"
    show (TApp t1 t2) = show t1 ++ "."++ show t2
    show (TSel t1 t2) = "(" ++ show t1 ++ "|" ++ show t2 ++ ")"
    show (TFun s) = show s
    show (TChar c) = "#"++[c]
    show (TRec r)  = showR r 
    show (TAnyLabel r e) = "~"++showR r++"("++show e++")"
    show (TAnyChar) = "."
    show (TStar t)  = "(" ++ show t ++ ")*"
    show (TNil)     = "$"
    
showR (TREmp)   = ""
showR (x)       = "{" ++ show x ++ "}"
 
data TName = TO String  -- Original Typename
           | TI Int     -- Newly introduced, but important because they are used in branches in the oroginal program
           | TV Int     -- Otherwise 
             deriving (Eq, Ord)
instance Show TName where
    show (TO a) = a
    show (TI i) = '#':show i
    show (TV i) = '@':show i

data TRec a = TREmp
            | TRCons (String,TExp a) (TRec a)
            | TRRest 
            | TRFun a -- Must be top
          deriving (Eq, Ord, Show)

instance Functor TRec where
    fmap f TREmp  = TREmp
    fmap f TRRest = TRRest
    fmap f (TRCons (k,t) r) = TRCons (k,fmap f t) (fmap f r)

divideDecl x = f x
    where f (TDecl t te:decls) = let (x,y) = f decls in (TDecl t te:x,y)
          f (FDecl g p e:decls) = let (x,y) = f decls in (x, FDecl g p e:y)
          f [] = ([],[])

{-
allTExp decls = f 0 $ allTExp' decls
    where allTExp' (TDecl t te:decls)  = (Right t,te):allTExp' decls
          allTExp' (FDecl f p e:decls) = 
              concatMap g p ++ allTExp' decls
          allTExp' [] = []
          g (PCon _ r e) = 
              h r ++ g e
          g (PApp t1 t2) = g t1 ++ g t2 
          g (PVar _  t) = [(Left (), t)]
          g (PNil) = []
          h _ = [] -- temporal 
          f n ((Right t,te):ts) = (TO t, fmap TO te):f n ts
          f n ((Left (),te):ts) = (TI n, fmap TO te):f (n+1) ts
          f n []                = []
-}