{-# OPTIONS -fglasgow-exts -XOverlappingInstances -XIncoherentInstances #-}

module Complementation where

import AST
import Data.List
import Control.Monad.State
import Debug.Trace
import Data.Function
import Data.Maybe
import Util

import Automaton



newtype CP a b = CP [(Int, 
                      (a, [IdPat a b], [IdExp a b], [([IdPat a b], IdExp a b)]))]
-- ^ Rule ID, (FunctionName, Patterns, Expressions, [(Pattern, Function Calls)])

importantStatesC (CP rules) =
    snub $ concatMap f rules
    where
      f (_,(_,ps,_,ws)) = concatMap fp ps ++ concatMap fw ws
      fp (PCon _ _ r e) = fp r ++ fp e
      fp (PApp _ e1 e2) = fp e1 ++ fp e2
      fp (PVar i _)     = [i] 
      fp (PChar _ _)    = []
      fp (PRec _ rs)    = concatMap (fp.snd) rs
      fp (PNil _)       = []
      fp (PFCon _ _ es) = concatMap fp es
      fw (ps,_)         = concatMap fp ps
    

instance (Show a, Show b) => ShowP (CP a b) where
    showP (CP ls) = 
        foldr (\a r -> a++"\n"++r) [] $
              map (\(i,(f,ps,es,ws)) ->
                       show i ++ ":\n" 
                                ++ show f ++ "(" ++ sP ps ++ ") = \n  "
                                ++ sE es ++ "\n" 
                                ++ g ws) ls
            where 
              sP [p] = showP p
              sP ps = "(" ++ implode ", " (map showP ps) ++ ")"
              sE [e] = showP e
              sE es = "(" ++ implode ", " (map showP es) ++ ")"
              g [] = []
              g ws = "  where\n"++(implode "\n" $ zipWith (\l r -> "    " ++ s l ++ " = " ++ r) ls rs)
                  where 
                    ls = map (sP.fst)    ws
                    rs = map (showP.snd) ws
                    maxls = foldr max 0 $ map length ls
                    s l = l ++ replicate (maxls - length l) ' '

split_to_funcs :: (Ord a, Eq a) => P a Int -> [(a, [(Int, [IdPat a Int], IdExp a Int)])]
split_to_funcs (P program) = 
    h $ groupBy ((==) `on` fname) $ sortBy (compare `on` fname) $ program
    where 
      fname (_,(f,_,_)) = f 
      h [] = []
      h ((x:xs):xss) = 
          (fname x, map g (x:xs)):h xss
      g (i,(_,ps,e)) = (i,ps,e)

{- |
 This function calculates overlap matrix from a given program.
 Concrely, for each function rules,
 this calculate there exists overlap or not.
 
 NOTE:
 Indices of a returned matrix is the pair of rule ids. 
-}
vertical_overlap_matrix :: (Ord a,Eq a) => P a Int -> Automaton (AState a Int) ->  [((Int,Int),Bool)]
vertical_overlap_matrix program am =
    evalState (concatMapM ol_matrix f_rules) ([],[])
    where
      f_rules = split_to_funcs program
      ol_matrix (f,decls) = 
          mapM g [ (i, i', expId e, expId e') | (i,_,e) <- decls, (i',_,e') <- decls ]
      g (rid,rid',j,j') = 
          do b <- checkOverlap (SE j) (SE j') am
             return $ ((rid,rid'), b)
          
{- |
 This function calculates horizontal overlaps,
 This information is used both in the injectivity analysis and in complement derivation.
-}
horizontal_overlap_matrix :: (Ord a,Eq a) => P a Int -> Automaton (AState a Int) -> [(Int,Bool)]
horizontal_overlap_matrix (P program) am = 
--    evalState (mapM h (appendsAll program)) ([],[])
    [ (i, evalState (checkOverlapH (SE j) (SE k) am) ([],[])) | (i,j,k) <- appendsAll program ]
    where
      appendsAll p = concatMap (\(_,(_,_,e)) -> appends e) p
      h (i,j,k) = 
          do b <- checkOverlapH (SE j) (SE k) am
             return $ (i,b)

appends (ECon _ c r e) = appends r ++ appends e
appends (ERec _ rs)    = concatMap (appends.snd) rs
appends (EApp i e1 e2) = (i,expId e1, expId e2):appends e1 ++ appends e2
appends _              = []

function_calls e = snub $ h e 
    where
      h (ECon _ _ r e) = h r ++ h e 
      h (EApp _ e1 e2) = h e1 ++ h e2
      h (ERec _ rs)    = concatMap (h.snd) rs
      h (EFun i f vs)  = [(i,f,vs)]
      h (EFCon _ _ es) = concatMap h es
--      h (EACon _ _ e es1 es2) = h e ++ concatMap h es1 ++ concatMap h es2
      h (ELen _ _ e) = h e
      h _ = []

varsP :: (Eq a, Ord a) => IdPat a b -> [a]
varsP p = snub $ h p
    where h (PCon _ _ r p) = h r ++ h p
          h (PFCon _ _ es) = concatMap h es
          h (PApp _ p1 p2) = h p1 ++ h p2
          h (PRec _ rs)    = concatMap (h.snd) rs
          h (PVar i v)     = [v]
          h _              = []

varsE :: (Eq a, Ord a) => IdExp a b -> [a]
varsE e = snub $ h e
    where h (ECon _ _ r p) = h r ++ h p
          h (EFCon _ _ es) = concatMap h es
          h (EApp _ p1 p2) = h p1 ++ h p2
          h (ERec _ rs)    = concatMap (h.snd) rs
          h (EVar i v)     = [v]
          h (EFun i f vs)  = vs
          h _              = []

-- True: Injective, False: Maybe non-Injective
check_injectivity :: (Ord a,Eq a) => P a Int -> [((Int,Int),Bool)] -> [(Int,Bool)] -> [a]
check_injectivity (P program) vols hols =
    map fst $ filter (snd) $ zip fnames (evalState (mapM ck fnames) [])
    where
      fnames  = snub $ map (\(_,(f,_,_)) -> f) program
      f_rules = split_to_funcs (P program)
      ck f = 
          do { proced <- get 
             ; if f `elem` proced then
                   return True
               else
                   if lostVar || hasVOL || hasHOL then
                       return $ False
                   else
                       do put (f:proced)
                          ls <- mapM ck fcalls
                          return $ and ls
             }
          where
            rules   = concat $ map snd $ filter (\(g,_) -> g==f) f_rules
            hasHOL  = any (\(_,_,e) -> any (\(i,j,k) -> fromJust$lookup i hols) (appends e)) rules
            hasVOL  = or [ fromJust$lookup (i,j) vols | (i,_,_) <- rules, (j,_,_) <- rules, i < j ]
            lostVar = any (\(_,ps,e) -> not $ null (snub (concatMap varsP ps) \\ varsE e)) rules
            fcalls  = concatMap (\(_,_,e) -> map (\(_,f,_) -> f) $ function_calls e) rules

-- TODO: If automaton only accept a singleton tree, replace it as a preprocess.

data Name a
    = Cmpl a
    | Inv  a
    | TInv a
    | Fwd  a
    | Bwd  a
    | V    a
    | VF   (Name a) Int  -- ^ Variable for Function Call
    | VE   Int           -- ^ Variable for Append Expression
    | VL   Int           -- ^ Variable for Length Expression
    | VI   Int           -- ^ Variable for System Use
    | VH   a Int         -- ^ Variable for a System Introduced Function for a Rule
    deriving (Ord, Eq)

instance Show (Name String) where
    show (Cmpl a) = a ++ "_c"
    show (Inv  a) = a ++ "_inv"
    show (TInv a) = a ++ "_Tinv"
    show (Fwd  a) = a
    show (Bwd  a) = a ++ "_b"
    show (V a)    = a
    show (VF f i) = show f ++ "__" ++ show i 
    show (VE i)   = "_cat_"++show i
    show (VL i)   = "_len_"++show i 
    show (VI i)   = "_var_"++show i
    show (VH a i) = "_f"++a++"_"++show i


instance Show a => Show (Name a) where
    show (Cmpl a) = show a ++ "_c"
    show (Inv  a) = show a ++ "_i"
    show (TInv a) = show a ++ "_Tinv"
    show (Fwd  a) = show a
    show (Bwd  a) = show a ++ "_b"
    show (V a)    = show a
    show (VF f i) = show f ++ "__" ++ show i 
    show (VE i)   = "_cat_"++show i
    show (VL i)   = "_len_"++show i 
    show (VI i)   = "_var_"++show i
    show (VH a i) = "_f"++show a++"_"++show i

uniq ::State Int Int
uniq = do { i <- get; put (i+1); return i }

-- | Returns Complement Function and Inversed Tupled Function
prepare_backward :: (Eq a, Ord a) => (P a Int, Automaton (AState a Int)) -> (CP (Name a) (AState a Int), CP (Name a) (AState a Int))
prepare_backward (program,am) = (introduce_where cmpl_prog, invt_prog)   
    where
      hols = horizontal_overlap_matrix program am
      vols = vertical_overlap_matrix program am
      injectives = check_injectivity program vols hols
      cmpl_prog  = complementation program vols hols injectives      
      invt_prog  = inverse_of_tupledfunction program cmpl_prog hols injectives 

complementation :: (Eq a,Ord a) => P a Int -> [((Int,Int),Bool)] -> [(Int,Bool)] -> [a] -> P (Name a) Int
complementation (P rules) vols hols injectives = optimization (P $ evalState (mapM cmpl rules) 0) vols
    where
      fnames     = snub $ map (\(_,(f,_,_)) -> f) rules
      cmpl (i,(f,p,e)) | f `elem` injectives = do { j <- uniq; return $ (i, (Cmpl f,map g p, EEps j)) }
                       | otherwise           = 
                           do { j <- uniq
                              ; e1 <- gather e
                              ; e2 <- lostvars p e 
                              ; return $ (i, (Cmpl f, map g p, EFCon j ("R",i) (e1++e2) )) }
          where lostvars p e = 
                    mapM (\v -> do { i <- uniq; return $ EVar i (V v) }) $ 
                                   (snub (concatMap varsP p) \\ varsE e)
                gather (ECon _ _ r e) = 
                    do r' <- gather r 
                       e' <- gather e
                       return $ r' ++ e'
                gather (EFun b f vs)  = 
                    if f `elem` injectives then
                        return []
                    else
                        do return $ [EFun b (Cmpl f) (map V vs)]
                gather (EApp b e1 e2) = 
                    if not$fromJust(lookup b hols) then 
                        do e1' <- gather e1 
                           e2' <- gather e2
                           return $ e1' ++ e2'
                    else
                        do e1' <- gather e1
                           e2' <- gather e2
                           i <- uniq 
                           j <- uniq 
                           return $ [ELen i b (conv e1)] ++ e1' ++ e2'
--                           return $ [EACon i b (ELen j (conv e1)) e1' e2']
                gather (EVar i v)  = return []                    
                gather (ERec i rs) =
                    do ls <- mapM (gather.snd) rs
                       return $ concat ls
                gather (EEps _)    = return []
                gather (EChar _ _) = return []
                conv (ECon i c r e) = ECon i c (conv r) (conv e)
                conv (EFun i f vs)  = EFun i (Fwd f) (map V vs)
                conv (EApp i e1 e2) = EApp i (conv e1) (conv e2)
                conv (EVar i v)     = EVar i (V v)
                conv (ERec i rs)    = ERec i (map (\(k,t) -> (k,conv t)) rs)
                conv (EChar i c)    = EChar i c
                conv (EEps i)       = EEps i
                g (PCon i c r p) = PCon i c (g r) (g p)
                g (PApp i p1 p2) = PApp i (g p1) (g p2)
                g (PVar i v)     = PVar i (V v)
                g (PChar i c)    = PChar i c
                g (PRec i rs)    = PRec i (map (\(k,p) -> (k,g p)) rs)
                g (PNil i)       = PNil i

    
{- |
 Remove constructor if the rule is independent
 in the sense that the corresponding rule in the forward transformation 
 does not overlap any other rule of the function.

 This is effective for zip like function.
 It is worth noting that we do not do constructor unification in this step.
 Constructor unification is less effective than constructor-removing.
 To make matters worth, it costs more time than constructor-removing.
 Moreover, look-aheads features allow users to cover the 
 situation.
-}                                                      
optimization :: (Eq a,Ord a) => P (Name a) Int -> [((Int,Int),Bool)] -> P (Name a) Int
optimization (P crules) vols =
    P $ sort $ combine_from_funcs $ map opt $ split_to_funcs (P crules)
    where
      combine_from_funcs [] = []
      combine_from_funcs ((f,xs):xss) =
          map (\(i,ps,e)->(i,(f,ps,e))) xs ++ combine_from_funcs xss
      opt :: (Name a, [(Int, [IdPat (Name a) Int], IdExp (Name a) Int)]) -> (Name a, [(Int, [IdPat (Name a) Int], IdExp (Name a) Int)])
      opt (f,xs) = (f,chk xs)
          where
            r_ids  = map (\(i,ps,e) -> i) xs            
            chk [] = []
            chk ((i,ps,EFCon j (r,k) [e]):ys) =
                if checkOL i r_ids then
                    (i,ps,EFCon j (r,k) [e]):chk ys
                else
                    (i,ps,e):chk ys
            chk (y:ys) = y:chk ys
            checkOL i r_ids = or [fromJust$lookup (i,j) vols|j <- r_ids,i/=j]

optimized (P crules) = snub $ map fst $ filter p_opt $ split_to_funcs (P crules)
    where
      p_opt (f,xs) = any (\(i,ps,e) -> chk e) xs
          where
            chk (EFCon _ _ _) = False
            chk _             = True
            

-- FIXME: variable name must be associated function call.
introduce_where :: Ord a => P (Name a) Int -> CP (Name a) (AState a Int)
introduce_where (P rules) = CP $ map iw rules
    where
      iw (i,(f,ps,e)) = (i,(f, map chVarP ps, [chVar e], ws++ws'))
          where
            fcalls = function_calls e
            lcalls = len_calls e
            ws  = map (\(i,f,vs) -> ([PVar (SE i) (VF f i)], EFun (SE (-i)) f vs)) fcalls
            ws' = map (\(ELen i b e)-> ([PVar (SE i) (VL b)], ELen (SE (-i)) (SE b) (chVar e))) lcalls
            len_calls (ECon _ _ r e) = len_calls r ++ len_calls e
            len_calls (EApp _ e1 e2) = len_calls e1 ++ len_calls e2
            len_calls (ERec _ rs)    = concatMap (len_calls.snd) rs
            len_calls (EFCon _ _ es) = concatMap len_calls es
--            len_calls (EACon _ _ e es1 es2) = len_calls e ++ concatMap len_calls es1 ++ concatMap len_calls es2
            len_calls (ELen i b e)   = [ELen i b e] ++ len_calls e 
            len_calls _ = []
            chVar (ECon i c r e) = ECon (SE i) c (chVar r) (chVar e)
            chVar (EApp i e1 e2) = EApp (SE i) (chVar e1) (chVar e2)
            chVar (EFun i f vs)  = EVar (SE i) (VF f i)
            chVar (EVar i v)     = EVar (SE i) v
            chVar (ERec i rs)    = ERec (SE i) (map (\(k,e) -> (k, chVar e)) rs)
            chVar (EChar i c)    = EChar (SE i) c
            chVar (EEps i)       = EEps (SE i)
            chVar (EFCon i c es)        = EFCon (SE i) c (map chVar es)
--            chVar (EACon i j e es1 es2) = EACon (SE i) (SE j) (chVar e) (map chVar es1) (map chVar es2)
            chVar (ELen  i b e)   = EVar (SE i) (VL b)
            chVarP (PCon i c r p) = PCon (SE i) c (chVarP r) (chVarP p)
            chVarP (PApp i p1 p2) = PApp (SE i) (chVarP p1) (chVarP p2)
            chVarP (PVar i v)     = PVar (SE i) v
            chVarP (PChar i c)    = PChar (SE i) c
            chVarP (PRec i rs)    = PRec (SE i) (map (\(k,p) -> (k, chVarP p)) rs)
            chVarP (PNil i)       = PNil (SE i)
            

change_names (P prog) = P $ map (\(i,(f,ps,e)) -> (i,(Fwd f,map chP ps, chE e))) prog
    where
      chP (PCon i c r p) = PCon i c (chP r) (chP p)
      chP (PApp i p1 p2) = PApp i (chP p1) (chP p2)
      chP (PVar i v)     = PVar i (V v)
      chP (PChar i c)    = PChar i c
      chP (PRec i rs)    = PRec i (map (\(k,t) -> (k,chP t)) rs)
      chP (PNil i)       = PNil i
      chE (ECon i c r p) = ECon i c (chE r) (chE p)
      chE (EApp i e1 e2) = EApp i (chE e1) (chE e2)
      chE (EFun i f vs)  = EFun i (Fwd f) (map V vs)
      chE (EVar i v)     = EVar i (V v)
      chE (ERec i rs)    = ERec i (map (\(k,t) -> (k,chE t)) rs)
      chE (EChar i c)    = EChar i c
      chE (EEps i)       = EEps i

                
inverse_of_tupledfunction :: (Eq a, Ord a) => P a Int -> P (Name a) Int -> [(Int,Bool)] -> [a] -> CP (Name a) (AState a Int)
inverse_of_tupledfunction (P fwd_prog) (P cmpl_prog) hols injectives = 
    CP $ concat $ zipWith f wfwd_prog wcmpl_prog
    where
      P fwd_prog'   = change_names (P fwd_prog)
      CP wfwd_prog  = introduce_where (P fwd_prog') 
      CP wcmpl_prog = introduce_where (P cmpl_prog) 
      optimized_f = optimized (P cmpl_prog)
      e2p' (ECon _ c r e) = PCon UD c (e2p' r) (e2p' e)
      e2p' (EVar _ v)     = PVar UD v
      e2p' (ERec _ rs)    = PRec UD $ map (\(k,p) -> (k,e2p' p)) rs
      e2p' (EFCon _ c es) = PFCon UD c $ map e2p' es
--      e2p' (EACon _ i e es1 es2) = PACon UD i (e2p' e) (map e2p' es1) (map e2p' es2)
      e2p' (EChar _ c)    = PChar UD c
      e2p' (EEps _)       = PNil UD
      f (i,(Fwd g,ps,[e],ws)) (i',(_,_,[e'],ws')) = -- FIXME: more Optimization
          if g `elem` injectives then
              let (_,ps',es',ws') = invert (Fwd g, ps, [e], ws)
              in [(i, (Inv g,ps',es',ws') ) ]--[(2*i, (Inv g, [PVar (expId e) (VI 0)], [EFun UD (VH g i) [VI 0]], [])), (2*i+1, (VH g i,ps',es',ws'))]
          else 
              [(i, (TInv g, [p,p'], map p2e ps, wsI))]
          where
            p'   = e2p' e' 
            p    = patternize e hols bind
            wsI  = gatherWhere e hols injectives (len_vars e') bind fargs 
            bind = concatMap binds ps
            fargs = [ ((h,i),vs) | ([PVar _ (VF h i)], EFun _ _ vs) <- ws]
      len_vars (ECon _ _ r e) = len_vars r ++ len_vars e
      len_vars (EFCon _ _ es) = concatMap len_vars es
      len_vars (ERec _ rs)    = concatMap (len_vars.snd) rs
      len_vars (EApp _ e1 e2) = len_vars e1 ++ len_vars e2
--      len_vars (EACon _ (SE i) (EVar _ (VL j)) es1 es2) = [(i,VL j)]++concatMap len_vars es1 ++ concatMap len_vars es2
      len_vars _ = []



invert :: Eq a => (Name a, [IdPat (Name a) (AState a Int)], [IdExp (Name a) (AState a Int)], [([IdPat (Name a) (AState a Int)], IdExp (Name a) (AState a Int))])
          -> (Name a, [IdPat (Name a) (AState a Int)], [IdExp (Name a) (AState a Int)], [([IdPat (Name a) (AState a Int)], IdExp (Name a) (AState a Int))])
invert (Fwd g,ps,[e],ws) = (Inv g, [e2p e], map p2e ps, ws')
    where
      bind = concatMap binds ps
      ws' = [ (map (e2p.v2e) vs, EFun UD (Inv f) [ v | PVar _ v <- ps ])  | (ps,EFun _ (Fwd f) vs) <- ws ]
          where v2e v = EVar (fromJust$lookup v bind) v

e2p (ECon i c r e) = PCon i c (e2p r) (e2p e)
e2p (EApp i e1 e2) = PApp i (e2p e1) (e2p e2)
e2p (EVar _ (VF (Fwd g) i)) = PVar (SF g) (VF (Fwd g) i)
e2p (EVar i v)     = PVar i v
e2p (ERec i rs)    = PRec i $ map (\(k,e) -> (k, e2p e)) rs
e2p (EChar i c)    = PChar i c
e2p (EEps i)       = PNil i


p2e (PCon _ c r p) = ECon UD c (p2e r) (p2e p)
p2e (PApp _ p1 p2) = EApp UD (p2e p1) (p2e p2)
p2e (PChar _ c)    = EChar UD c
p2e (PVar _ v)     = EVar  UD v
p2e (PRec _ rs)    = ERec  UD $ map (\(k,p) -> (k,p2e p))  rs
p2e (PNil _)       = EEps  UD
              
          
elim_redcalls (i,(f,ps,es,ws)) = (i,(f,ps,es,ws'))
    where
      ws' = step (concatMap varsE es) ws
      step vs ws =
          let (ws1,ws2) = partition (p vs) ws
          in case ws1 of 
               [] -> []
               _  -> ws1 ++ step (concatMap (\(_,e) -> varsE e) ws1) ws2
      p vs (ps,e) =
          let vps = concatMap varsP ps
          in any (\v -> v `elem` vps) vs 
                 

-- patternize :: IdExp (Name a) Int -> [(Int,Bool)] -> IdPat (Name a) Int
patternize :: Eq a => IdExp (Name a) (AState a Int) -> [(Int,Bool)] -> [(Name a,AState a Int)] -> IdPat (Name a) (AState a Int)
patternize exp hols bind = f exp
    where
      f (EEps i)                = PNil i
      f (EVar _ (VF (Fwd g) i)) = PVar (SF g) (VF (Fwd g) i)
      f (EVar i v)              = PVar (fromJust$lookup v bind) v
      f (ECon i c r e)          = PCon i c (f r) (f e)
      f (EApp (SE i) e1 e2) =
          if fromJust$lookup i hols then
              PVar (SE i) (VE i)
          else
              PApp (SE i) (f e1) (f e2)
      f (EChar i c) = PChar i c
      f (ERec i rs) =
          PRec i $ map (\(k,t) -> (k,f t)) rs
            
      
gatherWhere :: Eq a => 
               IdExp (Name a) (AState a Int) ->               
              [(Int,Bool)] ->                   
              [a] ->
              [(Int, Name a)] ->
              [(Name a,AState a Int)] -> 
              [((Name a,Int), [Name a])]-> 
              [([IdPat (Name a) (AState a Int)], IdExp (Name a) (AState a Int))]
gatherWhere exp hols injectives len_vars bind fargs = f exp
    where
      f (EEps _) = []
      f (EVar _ (VF (Fwd g) i)) =
          let args = fromJust $ lookup (Fwd g,i) fargs
              rets = map (\v -> PVar (fromJust $ lookup v bind) v) args
          in if g `elem` injectives then           
                 [(rets, EFun UD (Inv g) [VF (Fwd g) i])]
             else
                 [(rets, EFun UD (TInv g) [VF (Fwd g) i, VF (Cmpl g) i])]
      f (EVar _ v) = []
      f (ECon _ c r e) = f r ++ f e
      f (EApp (SE i) e1 e2) =
          if fromJust$lookup i hols then
              let p1   = patternize e1 hols bind 
                  p2   = patternize e2 hols bind
              in [ ([p1,p2], ESplitAt UD (VL i) (VE i) ) ] ++ f e1 ++ f e2
          else 
              f e1 ++ f e2
      f (EChar i c) = []
      f (ERec i rs) = concatMap (f.snd) rs

