{
module LangLex where
import Control.Monad.Fix
}

%wrapper "monad"

$alpha_or_number = [a-zA-Z0-9_]
$big_alpha       = [A-Z]
$small_alpha     = [a-z]

$safe_string_cont = [^ \\ \"]

@decimal = [1-9] [0-9]*
@ident   = _? $small_alpha $alpha_or_number* "'"*
@cident  = _? $big_alpha   $alpha_or_number* "'"*
@bident  = "<" ($big_alpha | $small_alpha) $alpha_or_number* ">"
@sident  = "$" ($big_alpha | $small_alpha) $alpha_or_number*
@keyword = "where" | "transformation" | "atomic" | "oneway" | "inversepair" | "with" | "data" | "protect" | "abort" | "backward" | "if" | "then" | "else" 
@string = ( $safe_string_cont* | "\\\"" )* 

tokens :-
<0> \" @string \" { mkTok $ \s -> TkString (tail (init s)) }
<0> "--" [^\n]*          ;
<0> "{-"                 { begin comment }
<0> $white+ ; -- We skip whitespaces
<0> "="                  { mkTok $ \s -> TkEq }
<0> "()"                 { mkTok $ \s -> TkUnit }
<0> "("                  { mkTok $ \s -> TkLPar }
<0> ")"                  { mkTok $ \s -> TkRPar }
<0> "{"                  { mkTok $ \s -> TkLBrace }
<0> "}"                  { mkTok $ \s -> TkRBrace }
<0> @keyword             { mkTok $ \s -> TkKey s }
-- <0> "where"              { mkTok $ \s -> TkWhere }
<0> "|"                  { mkTok $ \s -> TkPar }
<0> "::"                 { mkTok $ \s -> TkColons }
<0> ":"                  { mkTok $ \s -> TkColon }
<0> ";"                  { mkTok $ \s -> TkSemiColon }
<0> "->"                 { mkTok $ \s -> TkArr }
<0> "~"                  { mkTok $ \s -> TkTilde }
<0> ","                  { mkTok $ \s -> TkComma }
<0> "."                  { mkTok $ \s -> TkDot }
<0> "@"                  { mkTok $ \s -> TkAt }
<0> "["                  { mkTok $ \s -> TkLBracket }
<0> "]"                  { mkTok $ \s -> TkRBracket }
<0> "<"                  { mkTok $ \s -> TkLABracket }
<0> ">"                  { mkTok $ \s -> TkRABracket }
<0> "++"                 { mkTok $ \s -> TkAppend }
<0> "+"                  { mkTok $ \s -> TkPlus }
<0> "*"                  { mkTok $ \s -> TkAst }
<0> "-"                  { mkTok $ \s -> TkMinus }
<0> "/"                  { mkTok $ \s -> TkSlash }
<0> "^"                  { mkTok $ \s -> TkHat }
<0> "#"                  { mkTok $ \s -> TkSharp }
<0> "$"                  { mkTok $ \s -> TkDollar }
<0> @sident              { mkTok $ \s -> TkSIdent s }
<0> @ident               { mkTok $ \s -> TkIdent s }
<0> @cident              { mkTok $ \s -> TkCIdent s }
<0> @bident              { mkTok $ \s -> TkBIdent (tail (init s)) }
<0> "_"                  { mkTok $ \s -> TkUnderscore }
<0> @decimal             { mkTok $ \s -> TkInt (read s) }
<comment> "-}"           { begin 0 }
<comment> .              ;              
<comment> $white+        ;
<0> .                    { mkTok $ \s -> TkError s }
{
-- Token with Position
type PToken = ((Int,Int), Token)

-- pos :: PToken -> AlexPosn
-- pos (p,t) = p

body = snd

data Token 
    = TkEq            -- = 
    | TkUnit          -- ()
    | TkLPar          -- (
    | TkRPar          -- )
    | TkLBrace        -- {
    | TkRBrace        -- }
    | TkPar           -- |
    | TkColons        -- ::
    | TkColon         -- :
    | TkSemiColon     -- ;
    | TkArr           -- ->
    | TkComma         -- ,
    | TkDot           -- .
    | TkAt            -- @
    | TkTilde         -- ~
    | TkLBracket      -- [
    | TkRBracket      -- ]
    | TkLABracket     -- < 
    | TkRABracket     -- > 
    | TkAppend        -- ++
    | TkPlus          -- +
    | TkMinus         -- -
    | TkAst           -- *
    | TkSlash         -- /
    | TkSharp         -- #
    | TkDollar        -- $
    | TkUnderscore    -- _
    | TkHat           -- ^
    | TkKey    String -- 
    | TkWhere         -- where
    | TkSIdent  String -- variables
    | TkIdent  String -- variables
    | TkCIdent String -- Constructors 
    | TkBIdent String -- <test>
    | TkString String -- String
    | TkInt    Integer
    | TkEOF
    | TkError String 
      deriving (Eq,Ord,Show)

intValue (TkInt i) = i
strValue (TkIdent s)  = s
strValue (TkCIdent s) = s
strValue (TkBIdent s) = s
strValue (TkString s) = s
strValue (TkKey s)    = s
strValue _            = ""


-- instance Show Token where 
--     show TkEq         = "="
--     show TkLPar       = "("
--     show TkRPar       = ")"
--     show TkLBrace     = "{"
--     show TkRBrace     = "}"
--     show TkWhere      = "where"
--     show TkPar        = "|"
--     show TkColon      = ":"
--     show TkSemiColon  = ";"
--     show TkArr        = "->"
--     show TkComma      = ","
--     show TkAt         = "@"
--     show TkLBracket   = "["
--     show TkRBracket   = "]" 
--     show (TkIdent s)  = s
--     show (TkCIdent s) = s
--     show (TkFIdent s) = s
--     show (TkInt i)    = show i



mkTok f (AlexPn p n c,_,str) len = return $ ((n,c), f (take len str))
alexEOF = return $ (undefined, TkEOF)


scanTokens str 
    = runAlex str $ 
      fix (\loop ls -> do 
             { token@(p, tk) <- alexMonadScan;
               case tk of 
                 TkEOF -> return $ reverse ls 
                 _     -> loop $! (token:ls) }) []
}
