module Specialize where

import AST hiding (binds)
import Automaton
import Data.List
import Data.Maybe
import Data.Function (on)
import Control.Monad.State
import Util
import Value
import Debug.Trace 

type CAutomaton a = (Automaton a, [a])

data AmPat 
    = AmCon  String AmPat AmPat
    | AmApp  AmPat AmPat
    | AmVar  String ([Int],CAutomaton Int)
    | AmChar Char
    | AmRec  [(String,AmPat)]
    | AmNil  
      deriving (Show)

data AmExp
    = AECon  String AmExp AmExp
    | AEApp  AmExp AmExp
    | AEVar  String 
    | AEChar Char
    | AERec  [(String,AmExp)]
    | AEFun  String [([Int],CAutomaton Int)] [String]
    | AENil 
      deriving (Show)

funcalls x = f x
    where
      f (AECon _ e1 e2)  = f e1 ++ f e2
      f (AEApp e1 e2)    = f e1 ++ f e2
      f (AERec rs)       = concatMap (f.snd) rs
      f (AEFun f ams vs) = [AEFun f ams vs]
      f _                = []

-- ast2amast :: (Eq a, Ord a) => Automaton a -> [(Int,(String, [IdPat String a], IdExp String a))] -> [(String, [AmPat], AmExp)]
ast2amast am [] = []
ast2amast am ((i,(f,ps,e)):xs) = (f, ps', e2e e):ast2amast am xs
    where
      p2p = convPat am
      ps' = map p2p ps
      bs  = concatMap binds ps'
      e2e (ECon _ c e1 e2) = AECon  c (e2e e1) (e2e e2)
      e2e (EApp _ e1 e2)   = AEApp  (e2e e1) (e2e e2)
      e2e (EVar _ v)       = AEVar  v
      e2e (EChar _ c)      = AEChar c 
      e2e (ERec _ rs)      = AERec $ map (\(k,e) -> (k, e2e e)) rs
      e2e (EFun _ f vs)    = AEFun f (map (\v -> lookup' v bs) vs) vs
      e2e (EEps _)         = AENil
      lookup' b xs = case lookup b xs of
                       Just v  -> v
                       Nothing -> error "``lookup'' failed at ast2amast"

am2cam (Automaton transes) = (Automaton transes, [ x | (x,TEps) <- transes ])

-- convPat :: (Eq a, Ord a) => Automaton a -> IdPat String a -> AmPat 
convPat am (PNil _)       = AmNil
convPat am (PVar i v)     = AmVar v (is',am2cam am')
    where
      (is1,am1) = numerizeState [SE i] (Automaton (accessibleRules [SE i] (transitions am)))
      (is',am') = uncurry minimize $ uncurry numerizeState $ uncurry determinize $ (is1,am1)
convPat am (PChar _ c)    = AmChar c
convPat am (PCon _ c r p) = AmCon c (convPat am r) (convPat am p)
convPat am (PRec _ rs)    = AmRec $ map (\(k,p) -> (k,convPat am p)) rs
convPat am (PApp _ p1 p2) = AmApp (convPat am p1) (convPat am p2)

binds x = f x
    where f (AmCon _ p1 p2) = f p1 ++ f p2 
          f (AmApp p1 p2)   = f p1 ++ f p2
          f (AmVar s am)    = [(s,am)]
          f (AmRec rs)      = concatMap (f.snd) rs
          f _               = []


elemIndexBy eq a x = f 0 x
    where f n []  = Nothing
          f n  (x:xs) | eq a x     = Just n
                      | otherwise  = f (n+1) xs

lookupBy eq v [] = Nothing
lookupBy eq v ((a,b):x) 
    | eq v a    = {- trace (show v ++ show a ++ "\n\n")  $ -} Just b
    | otherwise = lookupBy eq v x
               


typeSpecialize am (P rules) = 
    preast2ast $ amast2preast $ replaceSingleton $ specialize (ast2amast am rules)

-- FIXME: eliminate redundunt epsilon
replaceSingleton rules = -- trace (show (rules)) $ 
    map h rules
    where
      h (f,ams,ps,e) = {-trace ("VS"++show (map fst  vs)) $ -} (f, ams',ps', repe e)
          where
            bs = concatMap binds ps
            vs = [ (x, fromJust v) | (x,(is,(am,qs))) <- bs, v <- [checkSingletonD is am], isJust v ]
            (ams',ps') = 
                case ams of
                  [] -> 
                      ([],map repp ps )
                  _  ->
                      unzip $ map (\(a,p) -> (a,repp p)) $ filter (\((is,(am,qs)),p)->not $ isJust (checkSingletonD is am)) $ zip ams ps 
            v2p (VEps)         = AmNil
            v2p (VChar c)      = AmChar c 
            v2p (VCon c v1 v2) = AmCon c (v2p v1) (v2p v2)
            v2p (VApp v1 v2)   = appp (v2p v1) (v2p v2)
            v2p (VRec rs)      = AmRec (map (\(k,v) -> (k,v2p v)) rs)
            v2e (VEps)         = AENil
            v2e (VChar c)      = AEChar c 
            v2e (VCon c v1 v2) = AECon c (v2e v1) (v2e v2)
            v2e (VApp v1 v2)   = appe (v2e v1) (v2e v2)
            v2e (VRec rs)      = AERec (map (\(k,v) -> (k,v2e v)) rs)
            repp AmNil           = AmNil
            repp (AmChar c)      = AmChar c 
            repp (AmCon c p1 p2) = AmCon c (repp p1) (repp p2)
            repp (AmApp p1 p2)   = appp (repp p1) (repp p2)
            repp (AmRec rs)      = AmRec (map (\(k,p) -> (k, repp p)) rs )
            repp (AmVar v a)     =
                case lookup v vs of
                  Just val -> v2p val
                  Nothing  -> AmVar v a
            repe AENil           = AENil
            repe (AEChar c)      = AEChar c 
            repe (AECon c e1 e2) = AECon c (repe e1) (repe e2)
            repe (AEApp e1 e2)   = appe (repe e1) (repe e2)
            repe (AERec rs)      = AERec (map (\(k,e) -> (k, repe e)) rs )
            repe (AEVar v)       =
                case lookup v vs of
                  Just val -> v2e val
                  Nothing  -> AEVar v
            repe (AEFun f ams args) = 
                AEFun f ams' args'
                where (ams', args') = unzip $ filter (\(_,v) -> not (v `elem` map fst vs)) $ zip ams args
            appp AmNil p2 = p2
            appp p1 AmNil = p1
            appp p1 p2    = AmApp p1 p2
            appe AENil e2 = e2
            appe e1 AENil = e1
            appe e1 e2    = AEApp e1 e2
                

amast2preast rules = 
    let (fdecls,(_,_,tdecls)) = runState (mapM g rules) (0,[],[]) 
    in (G $ snub tdecls, fdecls) 
    where
      uniq    = do { (i,v,g) <- get; put (i+1,v,g); return i }
      addTo n = do { (i,v,g) <- get; put (i+n,v,g) }
      addR (x,tr) = do { (i,v,g) <- get; put (i,v,(x,tr):g) }
      toI (is,(am,qs)) = 
          do { (i,ams,g) <- get;
             ; case lookupBy checkEquiv' (is,(am,qs)) ams of
                 Nothing ->
                     do { i <- uniq;
                        ; r <- fromAtoG i (numerizeState is am)
                        ; mapM addR r
                        ; (k,_,g') <- get
                        ; put (k,((is,(am,qs)),i):ams,g');
                        ; return i }
                 Just i -> 
                     return i
             }
      -- toI (is,(am,qs)) = 
      --     do { (i,ams,g) <- get;
      --        ; let v = toAutomaton (am,qs)
      --        ; case lookupBy (\(is,am) (is',am') -> checkEquiv is is' am am') (map L1 is,v) ams of
      --            Nothing ->
      --                do { i <- uniq;
      --                   ; r <- fromAtoG i (numerizeState (map L1 is) v)
      --                   ; mapM addR r
      --                   ; (k,_,g') <- get
      --                   ; put (k,((map L1 is,v),i):ams,g');
      --                   ; return i }
      --            Just i -> 
      --                return i
      --        }
      toNF f ams = 
          do ns <- mapM toI ams
             return $ f ++ show ns
      g (f, ams, ps, e) = 
          do { f'  <- toNF f ams
             ; ps' <- mapM toP ps
             ; e'  <- toE e
             ; return (f',ps',e') }
      toE (AECon c e1 e2) =
          do { e1' <- toE e1
             ; e2' <- toE e2
             ; return $ ECon' c e1' e2' }
      toE (AEApp e1 e2) =
          do { e1' <- toE e1
             ; e2' <- toE e2
             ; return $ EApp' e1' e2' }
      toE (AEVar v) = 
          return $ EVar' v 
      toE (AEChar c) =
          return $ EChar' c
      toE (AERec rs) =
          do { es' <- mapM toE (map snd rs)
             ; return $ ERec' (zip (map fst rs) es') }
      toE (AEFun f ams vs) = 
          do { f' <- toNF f ams 
             ; return $ EFun' f' vs }
      toE (AENil) =
          return $ EEps'
      toP (AmCon c p1 p2) = 
          do { p1' <- toP p1
             ; p2' <- toP p2 
             ; return $ PCon' c p1' p2' }
      toP (AmApp p1 p2) =
          do { p1' <- toP p1
             ; p2' <- toP p2
             ; return $ PApp' p1' p2'  }
      toP (AmChar c) =
          return $ PChar' c 
      toP (AmNil) =
          return $ PNil'
      toP (AmRec rs) =           
          do { ts' <- mapM toP $ map snd rs
             ; return $ PRec' (zip (map fst rs) ts') }
      toP (AmVar v (is,(am,qs))) =
          do { i <- toI (is,(am,qs))
             ; return $ PVar' v i }
      fromAtoG i (is,am) = -- trace (show (uncurry numerizeState $ determinize is am)) $  
          let mI = foldr max 0 (states am)
              vs = map (\j->(i,DS $ j+i+1)) is
              am' = fmap (+(i+1)) am
              gr  = map (\(x,tr) -> (x,h tr)) $ transitions am'
          in do { addTo (mI+2)
                ; return (vs++gr) }
          where
            h (TT (NLabel c) [q1,q2,q3]) = 
                DA (DT c (DS q1) (DS q2)) (DS q3)
            h (TT AnyLabel [q1,q2,q3]) =
                DA (DAT (DS q1) (DS q2)) (DS q3)
            h (TT AnyChar [q]) = 
                DA DAnyC (DS q)
            h (TT (NChar c) [q]) =
                DA (DC c) (DS q)
            h (TR rs RNothing) =
                DR (map (\(k,t) -> (k,DS t)) rs) DNothing
            h (TR rs (RWithout ng)) = 
                DR (map (\(k,t) -> (k,DS t)) rs) (DWithout ng)
            h (TEps) = DE
            h (TN v) = DS v
          
          
-- amast2ast rules = -- trace ("==============\n"++foldr (\a r -> a ++ "\n" ++ r) [] (map show rules) ++"\n\n") $ 
--     let (fdecls,(_,_,tdecls)) =  runState (f rules) (0,[],[])
--     in  {- trace ("VVVV"++show (G $ snub tdecls)++"EEE") $ -}  (G (concatMap (toG fdecls) fdecls++tdecls), P (zip [1..] fdecls))
--     where
--       f rules = mapM g rules
--       uniq    = do { (i,v,g) <- get; put (i+1,v,g); return i }
--       addTo n = do { (i,v,g) <- get; put (i+n,v,g) }
--       addR (x,tr) = do { (i,v,g) <- get; put (i,v,(x,tr):g) }
--       toI (is,(am,qs)) = 
--           do { (i,ams,g) <- get;
--              ; case lookupBy checkEquiv' (is,(am,qs)) ams of
--                  Nothing ->
--                      do { i <- uniq;
--                         ; r <- fromAtoG i (numerizeState is am )
--                         ; mapM addR r
--                         ; (i',ams',g') <- get
--                         ; put (i',(is,(am,qs)):ams',g');
--                         ; return i }
--                  Just i ->
--                      return i
--              }
--       toNF f ams = 
--           do ns <- mapM toI ams
--              return $ f ++ show ns
--       g (f, ams, ps, e) = 
--           do { f'  <- toNF f ams
--              ; ps' <- mapM toP ps
--              ; e'  <- toE e
--              ; return (f',ps',e') }
--       toE (AECon c e1 e2) =
--           do { e1' <- toE e1
--              ; e2' <- toE e2
--              ; i   <- uniq
--              ; return $ ECon i c e1' e2' }
--       toE (AEApp e1 e2) =
--           do { e1' <- toE e1
--              ; e2' <- toE e2
--              ; i   <- uniq
--              ; return $ EApp i e1' e2' }
--       toE (AEVar v) = 
--           do { i <- uniq
--              ; return $ EVar i v }
--       toE (AEChar c) =
--           do { i <- uniq
--              ; return $ EChar i c }
--       toE (AERec rs) =
--           do { es' <- mapM toE (map snd rs)
--              ; i  <- uniq
--              ; return $ ERec i (zip (map fst rs) es') }
--       toE (AEFun f ams vs) = 
--           do { f' <- toNF f ams 
--              ; i  <- uniq
--              ; return $ EFun i f' vs }
--       toE (AENil) =
--           do { i <- uniq
--              ; return $ EEps i  }
--       toP (AmCon c p1 p2) = 
--           do { p1' <- toP p1
--              ; p2' <- toP p2 
--              ; i   <- uniq
--              ; return $ PCon i c p1' p2' }
--       toP (AmApp p1 p2) =
--           do { p1' <- toP p1
--              ; p2' <- toP p2
--              ; i   <- uniq
--              ; return $ PApp i p1' p2'  }
--       toP (AmChar c) =
--           do { i <- uniq
--              ; return $ PChar i c }
--       toP (AmNil) =
--           do { i <- uniq
--              ; return $ PNil i }
--       toP (AmRec rs) =           
--           do { ts' <- mapM toP $ map snd rs
--              ; i <- uniq
--              ; return $ PRec i (zip (map fst rs) ts') }
--       toP (AmVar v (is,(am,qs))) =
--           do { i <- toI (is,(am,qs))
--              ; return $ PVar i v }
--       fromAtoG i (is,am) = -- trace (show (uncurry numerizeState $ determinize is am)) $  
--           let mI = foldr max 0 (states am)
--               vs = map (\j->(i,DS $ j+i+1)) is
--               am' = fmap (+(i+1)) am
--               gr  = map (\(x,tr) -> (x,h tr)) $ transitions am'
--           in do { addTo (mI+2)
--                 ; return (vs++gr) }
--           where
--             h (TT (NLabel c) [q1,q2,q3]) = 
--                 DA (DT c (DS q1) (DS q2)) (DS q3)
--             h (TT AnyLabel [q1,q2,q3]) =
--                 DA (DAT (DS q1) (DS q2)) (DS q3)
--             h (TT AnyChar [q]) = 
--                 DA DAnyC (DS q)
--             h (TT (NChar c) [q]) =
--                 DA (DC c) (DS q)
--             h (TR rs RNothing) =
--                 DR (map (\(k,t) -> (k,DS t)) rs) DNothing
--             h (TR rs (RWithout ng)) = 
--                 DR (map (\(k,t) -> (k,DS t)) rs) (DWithout ng)
--             h (TEps) = DE
--             h (TN v) = DS v

toAutomaton (Automaton transes, qs) =
    if qs == [ x | (x,TEps) <- transes ] then
        fmap L1 $ Automaton transes 
    else
        Automaton $ map (\q -> (L1 q, TEps)) qs ++ concatMap (g L1) transes
    where
      g k (q,tr) =
          case tr of
            TT (NLabel l) [q1,q2,q3] ->
                [(D q, TT (NLabel l) [D q1, D q2, D q3]), (k q, TT (NLabel l) [D q1, D q2, k q3])]
            TT AnyLabel [q1,q2,q3] ->
                [(D q, TT AnyLabel [D q1, D q2, D q3]), (k q, TT AnyLabel [D q1, D q2, k q3])]
            TT (NChar c) [q1] ->
                [(D q, TT (NChar c) [D q1]), (k q, TT (NChar c) [k q1])]
            TT AnyChar [q1] ->
                [(D q, TT AnyChar [D q1]), (k q, TT AnyChar [k q1])]
            TN q1 ->
                [(D q, TN (D q1)), (k q, TN (k q1))]
            TR rs ng -> 
                [(D q, TR (map (\(k,t) -> (k,D t)) rs) ng)]
            TEps -> 
                [(D q,TEps), (k q, TEps)]


sintersect t1 t2 = f (sort t1) (sort t2)
    where f (t1:ts1) (t2:ts2) 
            | t1 < t2   = f ts1 (t2:ts2)
            | t1 > t2   = f (t1:ts1) ts2
            | otherwise = t1:f ts1 ts2
          f _ _ = []

leaveAccessible (is,(am,qs)) =
    let am' = Automaton $ accessibleRules is $ transitions am
    in (is, (am', qs `sintersect` states am'))

spl :: AmPat -> ([Int],CAutomaton Int) -> [AmPat] 
spl (AmNil) (is,(Automaton transes,qs)) =
    if not (null(is `intersect` qs)) then 
        [AmNil]
    else
        []

spl (AmChar c) (is,(Automaton transes, qs)) = 
    if null $ intersect qs qs' then
        [AmChar c]
    else
        []
    where
      qs' = [ q' | (x,TT (NChar c) [q']) <- transes, x`elem`is ]

-- spl (AmVar v (ps,(am',qs'))) (is,(am,qs)) = -- (\s -> trace ("VAR:"++show ils') s) $ 
--     if all (\q -> checkEmpty q am'') ils' then
--         []
--     else
--         [AmVar v (ps'', (am'', sintersect qs'' (states am'')))]
--     where
--       js = [(Left q,Right y) | y <- ps, q <- is]
--       ls = [(Left x,Right y) | x<-qs, y<-qs' ]
--       (ps'',qs'') = splitAt (length js) ils'
--       (ils',am'') = numerizeState (js++ls) $ Automaton $ accessibleRules js $ transitions $ productMachine am am'

spl (AmVar v (is',(am',qs'))) (is,(am,qs)) = -- trace (show ((is,(am,qs)))) $ -- trace (">>>>\n"++show (is',(am',qs')) ++ show (is1,am1)++"<<<<<\n") $
    if all (\q -> checkEmpty q am3) is3 then
        []
    else
        [AmVar v (is3, (am3,qs3))]
    where
        (is1,am1) = uncurry minimize $ uncurry numerizeState $  {- uncurry determinize $ -} (map L1 is', toAutomaton (am',qs'))
        (is2,am2) = uncurry minimize $ uncurry numerizeState $  uncurry determinize $  (map L1 is , toAutomaton (am ,qs ))
        is_temp   = [ (Left q, Right q') | q <- is1, q' <- is2 ]
        (is3,am3) = uncurry minimize $ uncurry numerizeState $ (is_temp, Automaton $ accessibleRules is_temp $ transitions $ productMachine am1 am2 )
        qs3       = [ x | (x,TEps) <- transitions am3 ]
      

spl (AmCon c p1 p2) (is,(am,qs)) =
    [ AmCon c'' p1' p2' | 
      (x,TT c' [q1,q2,q3]) <- transitions am, x`elem`is, compatible c' (NLabel c), NLabel c'' <- [unify c' (NLabel c)], q3 `elem` qs, 
      p1' <- spl p1 (leaveAccessible ([q1],(am,qs'))), p2' <- spl p2 (leaveAccessible ([q2],(am,qs'))) ]
    where
      qs' = [ x | (x,TEps) <- transitions am ]
      
spl (AmRec rs) (is,(am,qs)) = 
    [ AmRec rs''
      | (x,TR rs' v) <- transitions am, x`elem`is, check rs rs' v,
        rs'' <- cart_prod [ [ (k,p')  | p' <- case lookup k rs' of 
                                                Just q' ->
                                                    spl p (leaveAccessible ([q'],(am,qs')))
                                                Nothing ->
                                                    case v of                             
                                                      RWithout vs | not (k `elem` vs) ->
                                                          [p]
                                                      _ ->
                                                          [] ] | (k,p) <- rs ] ]
    where
      qs' = [ x | (x,TEps) <- transitions am ]
      check rs rs' RNothing       = map fst rs == map fst rs'
      check rs rs' (RWithout vs)  = all (\k -> k `elem` ks' || (not (k `elem` vs))) (map fst rs)
          where ks' = map fst rs'
      cart_prod xs = foldr (\a r -> [p:q | p <- a, q <- r]) [[]] xs
                  
spl (AmApp (AmCon c p1 p2) p3) (is,(am,qs)) = 
    [ AmApp (AmCon c'' p1' p2') p3' | 
      (x,TT c' [q1,q2,q3]) <- transitions am, x`elem`is, compatible (NLabel c) c', NLabel c'' <- [unify (NLabel c) c'],
      p1' <- spl p1 (h ([q1],(am,qs'))), p2' <- spl p2 (h ([q2],(am,qs'))), p3' <- spl p3 (h ([q3],(am,qs))) ]
    where
      h   = leaveAccessible 
      qs' = [ x | (x,TEps) <- transitions am ]
            
spl (AmApp (AmChar c) p) (is,(am,qs)) = 
    [ AmApp (AmChar c'') p' | (x,TT c' [q1]) <- transitions am, x`elem`is, compatible (NChar c) c', NChar c'' <- [unify (NChar c) c'],
                            p' <- spl p (leaveAccessible ([q1],(am,qs)))]

spl (AmApp p1 p2) (is,(am,qs)) = -- trace ("App: \n" ++ show p1 ++ show p2 ++ "\n" ++ show (is,(am,qs)) ++ "\n" ++ show  [ AmApp p1' p2' | qend <- states am, p1' <- spl p1 (is,(am,[qend])), p2' <- spl p2 ([qend],(am,qs))]) $ 
    [ AmApp p1' p2' | qend <- states am, p1' <- spl p1 (is,(am,[qend])), p2' <- spl p2 (leaveAccessible ([qend],(am,qs)))]
          
    
inferencePattern p (is,(am,qs)) = -- trace (show p ++ "\n" ++ show (is,(am,qs)) ++ "\n" ++ show (spl p (is,(am,qs)))++"\n\n\n") $ 
    case spl p (is,(am,qs)) of
      []  -> trace "E" []
      [t] -> trace "S" [t]
      ts  -> trace "M" $ mergeFix ts
    where
      check p p' = and (zipWith checkEquiv' (map snd $ binds p) (map snd $ binds p'))
      mergeFix ts = 
          let ts' = merge ts 
          in if length ts' < length ts then
                 mergeFix ts'
             else
                 ts'
      merge [t]    = [t]
      merge (t:ts) = -- trace (show $ length (t:ts)) $ 
          case f t ts of
            Just ts' -> merge ts'
            Nothing  -> t:merge ts
       where
          f p [] = Nothing
          f p (p':ps) = 
              if length (filter (\(_,_,b) -> b==False) zs) == 0 then 
                  Just $ ps                     
              else if length (filter (\(_,_,b) -> b==False) zs) == 1 then -- trace ("!!!"++show (binds p) ++ "\n!!" ++show (map mg zs)) $ 
                  Just $ replaceVar p (map mg zs):ps
              else
                  do res <- f p ps 
                     return $ p':res
              where b1 = map snd $ binds p
                    b2 = map snd $ binds p'
                    zs = zipWith (\a b -> (a,b,checkEquiv' a b)) b1 b2
                    mg (a,b,True) = a
                    mg ((is1,(am1,qs1)), (is2,(am2,qs2)),False) = 
                        let (isL,(amL,qsL)) = (map Left is1, (fmap Left am1, map Left qs1))
                            (isR,(amR,qsR)) = (map Right is2, (fmap Right am2, map Right qs2))
                            am3             = Automaton $ transitions amL ++ transitions amR
                            is3             = isL ++ isR
                            qs3             = qsL ++ qsR
                            (vv,am)         = numerizeState (is3++qs3) am3
                            (is,qs)         = splitAt (length is3) vv
                        in {- (\s -> trace (">>>>\n"++show (is1,(am1)) ++ show (is2,(am2))++">>\n"++show s) s) $ -} simplifyC (is,(am,qs))

                        -- let states1 = snub ( states am1 ++ is1 )
                        --     states2 = snub ( states am2 ++ is2 )
                        --     am3     = productMachine am1 am2 
                        --     is3     = snub ([ (Left i, Right j) | i <- is1, j <- states2 ] ++ [ (Left i, Right j) | i <- states1, j <- is2 ] )
                        --     qs3     = snub ([ (Left i, Right j) | i <- qs1, j <- qs2 ])
                        --     (vv,am) = numerizeState (qs3++is3) am3
                        --     (is,qs) = splitAt (length qs3) vv
                        -- in 
      replaceVar p bs = fst $ repvar p bs
           where
             repvar (AmCon c r p) bs = 
                 let (r',bs')  = repvar r bs
                     (p',bs'') = repvar p bs'
                 in (AmCon c r' p', bs'')
             repvar (AmApp p1 p2) bs =
                 let (p1', bs')  = repvar p1 bs
                     (p2', bs'') = repvar p2 bs'
                 in (AmApp p1' p2', bs'')
             repvar (AmVar v _) (a:as) =
                 (AmVar v a, as)
             repvar (AmRec rs) bs =
                 let (rs',bs') = foldr (\(k,p) (r',bs)->
                                            let (p',bs') = repvar p bs 
                                            in ( (k,p'):r', bs' )) ([],bs) rs 
                 in (AmRec rs', bs')
             repVar AmNil bs = (AmNil,bs)

simplifyC (is,(am,qs)) = (is',(am',qs'))
    where
      (is',am') = uncurry minimize $ uncurry numerizeState $ uncurry determinize $ (map L1 is, toAutomaton (am,qs))
      qs'       = [ x | (x,TEps) <- transitions am' ]
         
simplify (is,(am,qs)) = (is',am')  
    where
      (is',am') = uncurry minimize $ uncurry numerizeState $ uncurry determinize $ (map L1 is, toAutomaton (am,qs))
      

specializeStep ams (f,ps,e) =
    [ (f,ps',  {- trace (show f ++ " " ++ show ps ++ "<<<>>> " ++ show ps') $ -} repl (concatMap binds ps') e) |
          ps' <- foldr (\(p, (q,(am,qs))) pss -> [ p':ps | p' <- inferencePattern p (q,(am,qs)), ps <- pss ]) [[]] (zip ps ams) ]
        where 
          repl bs (AECon c e1 e2)  = AECon c (repl bs e1) (repl bs e2)
          repl bs (AEApp e1 e2)    = AEApp (repl bs e1) (repl bs e2)
          repl bs (AEChar c)       = AEChar c
          repl bs (AERec rs)       = AERec $ map (\(k,t) -> (k, repl bs t)) rs
          repl bs (AENil)          = AENil
          repl bs (AEVar v)        = AEVar v
          repl bs (AEFun f ams vs) = AEFun f (map (\v-> lookup' v bs) vs) vs
          lookup' b xs = case lookup b xs of
                           Just v  -> v
                           Nothing -> error "``lookup'' failed at specializeStep"
                                     

elemBy f a [] = False
elemBy f a (x:xs)  
    | f a x     = True
    | otherwise = elemBy f a xs

specialize org_rules = 
    fx tl_rules next_init [] 
    where
      next = concatMap (\(_,_,_,e) -> funcalls e) 
      next_init = concatMap (\(_,_,e) -> funcalls e)  org_rules
      tl_rules  = let fs = snub $ map (\(AEFun f _ _)->f) next_init in
                  map (\(f,ps,e) -> (f,[],ps,e)) $ filter (\(f,_,_)->not(f`elem`fs)) org_rules
      rulesOf g = filter (\(f,_,_)->f==g) org_rules
      fx rules [] proced = trace ("Specialize Done!") rules
      fx rules (AEFun f ams _:nex) proced 
          | any (equiv (f,ams)) proced = fx rules nex proced
          | otherwise = -- trace ("TRY:"++show (f,ams))$
              let rs  = nubBy check [ (f',ams,ps',e') | (_,ps,e) <- rulesOf f, (f',ps',e') <- specializeStep ams (f,ps,e) ]
              in -- trace ("GEN:"++show rs1++"____") $  
                fx (rs++rules) (nex++next rs) ((f,ams):proced)
      equiv (f,ams) (f',ams') = 
          f==f' && and (zipWith 
                        (\(is,(am,qs)) (is',(am',qs')) -> checkEquivD is is' am am')
                        ams ams')
      check (f,ams,ps,e) (f',ams',ps',e') =
          f == f' && and (zipWith shapeEq ps ps') && and (zipWith checkEquiv' (map snd $ concatMap binds ps) (map snd $ concatMap binds ps'))


checkEquiv' (is,(am,qs)) (is',(am',qs')) =
    if qs == snub [ x | (x,TEps) <- transitions am ] && qs' == snub [ x | (x,TEps) <- transitions am' ] then 
        checkEquivD is is' am am'
    else -- trace (show (is,(am,qs)) ++ show (is',(am',qs'))) $ 
        checkEquiv (map L1 is) (map L1 is') (toAutomaton (am,qs)) (toAutomaton (am',qs'))

shapeEq :: AmPat -> AmPat -> Bool
shapeEq (AmCon c r t) (AmCon c' r' t') = c == c' && shapeEq r r' && shapeEq t t' 
shapeEq (AmApp t1 t2) (AmApp t1' t2')  = shapeEq t1 t1' && shapeEq t2 t2'
shapeEq (AmVar v1 _) (AmVar v2 _) = v1 == v2
shapeEq (AmChar c) (AmChar c') = c == c'
shapeEq (AmRec rs) (AmRec rs') = and (zipWith (==) (map fst rs) (map fst rs')) && and (zipWith shapeEq (map snd rs) (map snd rs'))
shapeEq (AmNil) (AmNil) = True
shapeEq _ _ = False